/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/audio.c: Audio device interface.
 */

#include "common.h"
#include "audio.h"

#include <alsa/asoundlib.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* The ALSA sound device handle.  NULL means the device is not in use. */
static snd_pcm_t *pcm_handle = NULL;

/* Number of bytes per audio sample. */
static unsigned int bytes_per_sample;
/* Number of samples in the internal audio buffer. */
static unsigned int hw_buffer_size;
/* Number of samples in the optimum transfer block size. */
static unsigned int hw_period_size;

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/**
 * audio_open:  Open the audio device for recording (output == 0) or
 * playback (output != 0).
 *
 * [Parameters]
 *       output: True (nonzero) to open the audio device for playback,
 *                  false (zero) to open the audio device for recording
 *         bits: Audio data resolution in bits (8 or 16)
 *     channels: Number of audio channels (at least 1)
 *        srate: Sampling rate of the audio data (at least 1)
 * [Return value]
 *     Nonzero on success, zero on error
 */
int audio_open(int output, unsigned int bits, unsigned int channels,
               unsigned int srate)
{
    PRECOND(bits == 8 || bits == 16, return 0);
    PRECOND(channels >= 1, return 0);
    PRECOND(srate >= 1, return 0);

    int result;
    const int stream =
        output ? SND_PCM_STREAM_PLAYBACK : SND_PCM_STREAM_CAPTURE;
    const int format = bits==8 ? SND_PCM_FORMAT_U8 : SND_PCM_FORMAT_S16_LE;
    snd_pcm_hw_params_t *hwparams;
    snd_pcm_hw_params_alloca(&hwparams);
    snd_pcm_sw_params_t *swparams;
    snd_pcm_sw_params_alloca(&swparams);

    if ((result = snd_pcm_open(&pcm_handle, "default", stream, 0)) < 0) {
        DMSG("snd_pcm_open(): %s", snd_strerror(result));
        goto error_return;
    }
    if ((result = snd_pcm_hw_params_any(pcm_handle, hwparams)) < 0) {
        DMSG("snd_pcm_hw_params_any(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params_set_access(pcm_handle, hwparams, SND_PCM_ACCESS_RW_INTERLEAVED)) < 0) {
        DMSG("snd_pcm_hw_params_set_access(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params_set_format(pcm_handle, hwparams, format)) < 0) {
        DMSG("snd_pcm_hw_params_set_format(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params_set_channels(pcm_handle, hwparams, channels)) < 0) {
        DMSG("snd_pcm_hw_params_set_channels(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params_set_rate(pcm_handle, hwparams, srate, 0)) < 0) {
        DMSG("snd_pcm_hw_params_set_rate(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params(pcm_handle, hwparams)) < 0) {
        DMSG("snd_pcm_hw_params(): %s", snd_strerror(result));
        goto error_close;
    }

    snd_pcm_uframes_t buffer_size, period_size;
    if ((result = snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_size)) < 0) {
        DMSG("snd_pcm_hw_params_get_buffer_size(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_hw_params_get_period_size(hwparams, &period_size, NULL)) < 0) {
        DMSG("snd_pcm_hw_params_get_period_size(): %s", snd_strerror(result));
        goto error_close;
    }

    if ((result = snd_pcm_sw_params_current(pcm_handle, swparams)) < 0) {
        DMSG("snd_pcm_sw_params_current(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_sw_params_set_avail_min(pcm_handle, swparams, period_size)) < 0) {
        DMSG("snd_pcm_sw_params_set_avail_min(): %s", snd_strerror(result));
        goto error_close;
    }
    if ((result = snd_pcm_sw_params(pcm_handle, swparams)) < 0) {
        DMSG("snd_pcm_sw_params(): %s", snd_strerror(result));
        goto error_close;
    }

    hw_buffer_size = buffer_size;
    hw_period_size = period_size;
    bytes_per_sample = bits/8 * channels;
    return 1;

  error_close:
    snd_pcm_close(pcm_handle);
  error_return:
    pcm_handle = NULL;
    return 0;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_close:  Close the audio device.  Does nothing if the audio device
 * was not open.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     None
 */
void audio_close(void)
{
    if (pcm_handle) {
        snd_pcm_close(pcm_handle);
        pcm_handle = NULL;
    }
}

/*************************************************************************/

/**
 * audio_buffer_size:  Return the optimum buffer size for reading or
 * writing audio data.  The return value is undefined if the audio device
 * is not open.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Optimum buffer size, in samples
 */
unsigned int audio_buffer_size(void)
{
    return hw_period_size;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_playable:  Return whether audio data can currently be played
 * through the audio device without blocking.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Nonzero if audio data can be played, else zero
 */
int audio_playable(void)
{
    PRECOND(pcm_handle != NULL, return 0);

    int result;
    do {
        result = snd_pcm_wait(pcm_handle, 0);
    } while (result < 0 && snd_pcm_recover(pcm_handle, result, 0) == 0);
    return result == 1;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_samples_buffered:  Return the number of samples buffered for
 * playback.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Number of samples buffered for playback
 */
unsigned int audio_samples_buffered(void)
{
    if (pcm_handle) {
        int n = snd_pcm_avail(pcm_handle);
        if (n >= 0) {
            return hw_buffer_size - n;
        }
    }
    return 0;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_samples_available:  Return the number of recorded samples
 * available for reading.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Number of samples available for reading
 */
unsigned int audio_samples_available(void)
{
    if (pcm_handle) {
        int n = snd_pcm_avail(pcm_handle);
        if (n >= 0) {
            return n;
        }
    }
    return 0;
}

/*************************************************************************/

/**
 * audio_play:  Play the given audio data through the audio device.
 *
 * [Parameters]
 *      data: Data to play
 *     count: Number of samples in data
 * [Return value]
 *     Nonzero on success, zero on error
 */
int audio_play(const void *data, uint64_t count)
{
    PRECOND(pcm_handle != NULL, return 0);
    PRECOND(data != NULL, return 0);

    const uint8_t *ptr = (const uint8_t *)data;
    while (count > 0) {
        unsigned int this_count = (unsigned int)ubound(count, hw_buffer_size);
        ssize_t result;
        while ((result = snd_pcm_writei(pcm_handle, ptr, this_count)) < 0) {
            result = snd_pcm_recover(pcm_handle, result, 0);
            if (result < 0) {
                DMSG("snd_pcm_writei(): %s", snd_strerror(result));
                return 0;
            }
        }
        ptr += result * bytes_per_sample;
        count -= result;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_record:  Record audio data from the audio device.
 *
 * [Parameters]
 *       buf: Buffer into which data should be written
 *     count: Number of samples to record
 * [Return value]
 *     Nonzero on success, zero on error
 * [Notes]
 *     An error return with errno set to EPIPE indicates that a device
 *     buffer overrun occurred, and some data was lost.  Further recording
 *     is still possible in this case.
 */
int audio_record(void *data, uint64_t count)
{
    PRECOND(pcm_handle != NULL, return 0);
    PRECOND(data != NULL, return 0);

    uint8_t *ptr = (uint8_t *)data;
    while (count > 0) {
        unsigned int this_count = (unsigned int)ubound(count, hw_buffer_size);
        ssize_t result;
        while ((result = snd_pcm_readi(pcm_handle, ptr, this_count)) < 0) {
            errno = -result;
            if (errno == EPIPE) {
                DMSG("Overrun detected, aborting read");
                result = snd_pcm_prepare(pcm_handle);
                if (result < 0) {
                    DMSG("snd_pcm_readi(): overrun recovery failed: %s",
                         snd_strerror(result));
                }
                return 0;
            } else {
                result = snd_pcm_recover(pcm_handle, result, 0);
                if (result < 0) {
                    DMSG("snd_pcm_readi(): recover: %s", snd_strerror(result));
                    return 0;
                }
            }
        }
        ptr += result * bytes_per_sample;
        count -= result;
    }

    return 1;
}

/*-----------------------------------------------------------------------*/

/**
 * audio_flush:  Flush all buffered playback data.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     None
 */
void audio_flush(void)
{
    PRECOND(pcm_handle != NULL, return);

    snd_pcm_nonblock(pcm_handle, 0);
    snd_pcm_drain(pcm_handle);
    snd_pcm_nonblock(pcm_handle, 1);
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
