/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/audio.h: Header for audio device interface.
 */

#ifndef AUDIO_H
#define AUDIO_H

/*************************************************************************/

/**
 * audio_open:  Open the audio device for recording (output == 0) or
 * playback (output != 0).
 *
 * [Parameters]
 *       output: True (nonzero) to open the audio device for playback,
 *                  false (zero) to open the audio device for recording
 *         bits: Audio data resolution in bits (8 or 16)
 *     channels: Number of audio channels (at least 1)
 *        srate: Sampling rate of the audio data (at least 1)
 * [Return value]
 *     Nonzero on success, zero on error
 */
extern int audio_open(int output, unsigned int bits, unsigned int channels,
                      unsigned int srate);

/**
 * audio_close:  Close the audio device.  Does nothing if the audio device
 * was not open.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     None
 */
extern void audio_close(void);

/*----------------------------------*/

/**
 * audio_buffer_size:  Return the optimum buffer size for reading or
 * writing audio data.  The return value is undefined if the audio device
 * is not open.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Optimum buffer size, in samples
 */
extern unsigned int audio_buffer_size(void);

/**
 * audio_playable:  Return whether audio data can currently be played
 * through the audio device without blocking.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Nonzero if audio data can be played, else zero
 */
extern int audio_playable(void);

/**
 * audio_samples_buffered:  Return the number of samples buffered for
 * playback.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Number of samples buffered for playback
 */
extern unsigned int audio_samples_buffered(void);

/**
 * audio_samples_available:  Return the number of recorded samples
 * available for reading.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Number of samples available for reading
 */
extern unsigned int audio_samples_available(void);

/*----------------------------------*/

/**
 * audio_play:  Play the given audio data through the audio device.
 *
 * [Parameters]
 *      data: Data to play
 *     count: Number of samples in data
 * [Return value]
 *     Nonzero on success, zero on error
 */
extern int audio_play(const void *data, uint64_t count);

/**
 * audio_record:  Record audio data from the audio device.
 *
 * [Parameters]
 *       buf: Buffer into which data should be written
 *     count: Number of samples to record
 * [Return value]
 *     Nonzero on success, zero on error
 * [Notes]
 *     An error return with errno set to EPIPE indicates that a device
 *     buffer overrun occurred, and some data was lost.  Further recording
 *     is still possible in this case.
 */
extern int audio_record(void *data, uint64_t count);

/**
 * audio_flush:  Flush all buffered playback data.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     None
 */
extern void audio_flush(void);

/*************************************************************************/

#endif  // AUDIO_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
