/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/canvas.c: Wave file canvas window management.
 */

#include "common.h"
#include "canvas.h"
#include "manager.h"
#include "ui.h"
#include "wave.h"

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glib.h>  // For g_path_get_basename()

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* Data structure for canvas windows (opaque to callers). */

struct CanvasWindow_ {
    /* Wave file with which this canvas window is associated. */
    WaveFile *wave;

    /* The GtkWindow for this canvas. */
    GtkWidget *window;
    /* The GtkDrawingArea on which we draw the waveform. */
    GtkWidget *canvas;

    /* The current display starting position, in samples. */
    uint64_t display_start;
    /* The last clicked position (used by dragging code). */
    uint64_t clicked_position;
};

/*************************************************************************/

/* Widget callback declarations. */

static gboolean cb_event_configure(GtkWidget *widget, GdkEvent *event,
                                   CanvasWindow *canvas);
static gboolean cb_event_delete(GtkWidget *widget, GdkEvent *event,
                                CanvasWindow *canvas);
static gboolean cb_event_expose(GtkWidget *widget, GdkEventExpose *event,
                                CanvasWindow *canvas);
static gboolean cb_event_focus_in(GtkWidget *widget, GdkEvent *event,
                                  CanvasWindow *canvas);
static gboolean cb_event_key_press(GtkWidget *widget, GdkEventKey *event,
                                   CanvasWindow *canvas);
static gboolean cb_event_button_press(GtkWidget *widget, GdkEventButton *event,
                                      CanvasWindow *canvas);
static gboolean cb_event_motion(GtkWidget *widget, GdkEventMotion *event,
                                CanvasWindow *canvas);
static gboolean cb_event_scroll(GtkWidget *widget, GdkEventScroll *event,
                                CanvasWindow *canvas);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/**
 * canvas_create:  Create a new canvas window for the given wave file.
 *
 * [Parameters]
 *     wave: Wave file for which to create a canvas window
 * [Return value]
 *     New canvas window, or NULL on error
 */
CanvasWindow *canvas_create(WaveFile *wave)
{
    /* Create and initialize the data structure itself. */
    CanvasWindow *canvas = malloc(sizeof(*canvas));
    if (!canvas) {
        DMSG("Out of memory");
        return NULL;
    }
    canvas->wave = wave;
    canvas->display_start = 0;
    canvas->clicked_position = 0;

    /* Create the GTK+ toplevel window for the canvas.  The window title is
     * set to the last (filename) component of the wave file's pathname. */
    canvas->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    if (!canvas->window) {
        DMSG("Out of memory");
        canvas_destroy(canvas);
        return NULL;
    }

    /* Create the drawing area for the canvas along with a decorative frame. */
    GtkWidget *frame = gtk_frame_new(NULL);
    if (!frame) {
        DMSG("Out of memory");
        canvas_destroy(canvas);
        return NULL;
    }
    gtk_container_set_border_width(GTK_CONTAINER(frame), 2);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
    gtk_container_add(GTK_CONTAINER(canvas->window), frame);
    canvas->canvas = gtk_drawing_area_new();
    if (!canvas->canvas) {
        DMSG("Out of memory");
        canvas_destroy(canvas);
        return NULL;
    }
    gtk_widget_set_can_focus(canvas->canvas, TRUE);
    gtk_container_add(GTK_CONTAINER(frame), canvas->canvas);

    /* Set reasonable size parameters for the canvas. */
    gtk_widget_set_size_request(canvas->canvas, 300, 100);
    GdkGeometry geometry = {
        .min_width = -1, .min_height = -1,  // -1 means "use the size request".
        .base_width = -1, .base_height = -1,
    };
    gtk_window_set_geometry_hints(GTK_WINDOW(canvas->window), canvas->canvas,
                                  &geometry, GDK_HINT_MIN_SIZE);
    /* +8 here accounts for the width of the frame border. */
    gtk_window_set_default_size(GTK_WINDOW(canvas->window), 700+8, 450+8);

    /* Connect signals to the drawing area. */
    gtk_widget_add_events(canvas->canvas, GDK_STRUCTURE_MASK
                                        | GDK_EXPOSURE_MASK
                                        | GDK_FOCUS_CHANGE_MASK
                                        | GDK_KEY_PRESS_MASK
                                        | GDK_BUTTON_PRESS_MASK
                                        | GDK_BUTTON_MOTION_MASK
                                        | GDK_SCROLL_MASK);
    g_signal_connect(canvas->canvas, "configure-event",
                     G_CALLBACK(cb_event_configure), canvas);
    // This is intentionally canvas->window; only windows get delete events.
    g_signal_connect(canvas->window, "delete-event",
                     G_CALLBACK(cb_event_delete), canvas);
    g_signal_connect(canvas->canvas, "expose-event",
                     G_CALLBACK(cb_event_expose), canvas);
    g_signal_connect(canvas->canvas, "focus-in-event",
                     G_CALLBACK(cb_event_focus_in), canvas);
    g_signal_connect(canvas->canvas, "focus-out-event",
                     G_CALLBACK(gtk_true), canvas);
    g_signal_connect(canvas->canvas, "key-press-event",
                     G_CALLBACK(cb_event_key_press), canvas);
    g_signal_connect(canvas->canvas, "button-press-event",
                     G_CALLBACK(cb_event_button_press), canvas);
    g_signal_connect(canvas->canvas, "motion-notify-event",
                     G_CALLBACK(cb_event_motion), canvas);
    g_signal_connect(canvas->canvas, "scroll-event",
                     G_CALLBACK(cb_event_scroll), canvas);

    /* Display the canvas window. */
    gtk_widget_show_all(canvas->window);

    /* Call refresh() to set the window title. */
    canvas_refresh(canvas, 0);

    return canvas;
}

/*-----------------------------------------------------------------------*/

/**
 * canvas_destroy:  Destroy a canvas window and all associated resources.
 * Does nothing if canvas == NULL.
 *
 * [Parameters]
 *     canvas: Canvas window to destroy
 * [Return value]
 *     None
 */
void canvas_destroy(CanvasWindow *canvas)
{
    if (canvas) {
        gtk_widget_destroy(canvas->window);  // Also destroys all subwidgets.
        free(canvas);
    }
}

/*************************************************************************/

/**
 * canvas_width:  Return the width of a canvas window.
 *
 * [Parameters]
 *     canvas: Canvas window for which to return width
 * [Return value]
 *     Window width, in pixels
 */
unsigned int canvas_width(CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return 1);
    GtkAllocation allocation;
    gtk_widget_get_allocation(canvas->canvas, &allocation);
    return allocation.width;
}

/*************************************************************************/

/**
 * canvas_refresh:  Refresh the contents of the canvas window, optionally
 * updating the display position.
 *
 * [Parameters]
 *         canvas: Canvas window to refresh
 *     reposition: Nonzero to reposition the wave data display position,
 *                    zero to keep the current display position
 * [Return value]
 *     None
 */
void canvas_refresh(CanvasWindow *canvas, int reposition)
{
    PRECOND(canvas != NULL, return);

    GtkAllocation allocation;
    gtk_widget_get_allocation(canvas->canvas, &allocation);

    /* Set the window title based on the current filename. */
    char *filename;
    if (wave_path(canvas->wave)) {
        filename = g_path_get_basename(wave_path(canvas->wave));
    } else {
        filename = strdup("<New file>");
    }
    gtk_window_set_title(GTK_WINDOW(canvas->window),
                         filename ? filename : "<Out of memory>");
    free(filename);

    /* Move the displayed portion of the wave data, if requested. */
    if (reposition) {
        const uint64_t half_window =
            wave_zoom(canvas->wave) * (allocation.width/2);
        if (wave_select_len(canvas->wave) > 0) {
            /* Position the window so the end of the selection is in the
             * middle of the window, if possible. */
            const uint64_t selection_end =
                wave_position(canvas->wave) + wave_select_len(canvas->wave);
            if (selection_end < half_window) {
                canvas->display_start = 0;
            } else {
                canvas->display_start = selection_end - half_window;
            }
        } else {
            /* Position the window so the current cursor position is at the
             * left edge of the canvas, but leave at least half a window's
             * worth of data visible (unless the entire wave file is
             * smaller than half a window). */
            uint64_t limit;
            if (wave_samples(canvas->wave) < half_window) {
                limit = 0;
            } else {
                limit = wave_samples(canvas->wave) - half_window;
            }
            canvas->display_start = ubound(wave_position(canvas->wave), limit);
        }
    }  // if (reposition)

    /* Generate an expose event to cause the refresh. */
    allocation.x = allocation.y = 0;
    gdk_window_invalidate_rect(gtk_widget_get_window(canvas->canvas),
                               &allocation, FALSE);
}

/*************************************************************************/
/*************************** Widget callbacks ****************************/
/*************************************************************************/

/**
 * cb_event_*:  Callbacks for GDK events on canvas windows.
 *
 * [Parameters]
 *     widget: Relevant widget (GtkDrawingArea, except for the delete event)
 *      event: Event data
 *     canvas: Canvas window data structure
 * [Return value]
 *     TRUE if further GTK+ processing should be halted, else FALSE
 */

static gboolean cb_event_configure(GtkWidget *widget, GdkEvent *event,
                                   CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    /* Update the UI first, since the maximum zoom factor is calculated
     * from the canvas width (so if the window has been expanded, the
     * zoom factor may change). */
    ui_update();

    canvas_refresh(canvas, 0);
    return FALSE;
}

/*----------------------------------*/

static gboolean cb_event_delete(GtkWidget *widget, GdkEvent *event,
                                CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    ui_close_wave(canvas->wave);
    return TRUE;  // Don't destroy the window immediately.
}

/*----------------------------------*/

static gboolean cb_event_expose(GtkWidget *widget, GdkEventExpose *event,
                                CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    /* Create a graphics context for drawing. */
    GdkDrawable *drawable = GDK_DRAWABLE(gtk_widget_get_window(widget));
    GdkGC *gc = gdk_gc_new(drawable);
    if (!gc) {
        DMSG("Failed to create GC");
        return FALSE;
    }

    /* Retrieve the size of the drawing area. */
    int width, height;
    gdk_drawable_get_size(drawable, &width, &height);

    /* Start out by painting the background.  We use the default GTK color
     * rather than white to differentiate this window from the main canvas
     * window.  (If the default background color is white, there's no
     * obvious difference in the first place, so we lose.) */
    gtk_style_apply_default_background(gtk_widget_get_style(widget),
                                       GDK_WINDOW(drawable), TRUE,
                                       GTK_STATE_NORMAL, &event->area,
                                       0, 0, width, height);

    /* Draw data for the wave file. */

    const uint64_t zoom = wave_zoom(canvas->wave);  // Just for convenience
    const uint64_t display_start = canvas->display_start;
    const uint64_t len = ubound(width * zoom,
                                wave_samples(canvas->wave) - display_start);
    float *data = wave_get_data(canvas->wave, display_start, len);
    const float *ptr = data;
    if (data) {
        const unsigned int channels = wave_channels(canvas->wave);
        const unsigned int channel_height = height / channels;

        /* Draw the actual wave data.  To avoid slowing down overly much at
         * high zoom levels, we only draw lines between the lowest and
         * highest points on each channel when the zoom level is greater
         * than 10. */
        unsigned int x;
        for (x = event->area.x;
             x < event->area.x + event->area.width && x * zoom < len;
             x++, ptr += channels * zoom
        ) {
            const uint64_t this_len = ubound(zoom, len - x * zoom);
            unsigned int ch;
            for (ch = 0; ch < channels; ch++) {
                const float *this_ptr = &ptr[ch];
                const unsigned int ybase = (ch+1) * channel_height - 1;
                if (zoom <= 10) {
                    uint64_t i;
                    for (i = 0; i < this_len; i++, this_ptr += channels) {
                        const unsigned int sample = bound((int)roundf(*this_ptr * 32768), -32768, 32767) + 0x8000;
                        const unsigned int y =
                            ybase - (sample * channel_height) / 65536;
                        gdk_gc_set_foreground(gc, wave_get_channel_color(canvas->wave, ch));
                        gdk_draw_point(drawable, gc, x, y);
                    }
                } else {
                    uint16_t low = 0xFFFF, high = 0;
                    uint64_t i;
                    for (i = 0; i < this_len; i++, this_ptr += channels) {
                        const unsigned int sample = bound((int)roundf(*this_ptr * 32768), -32768, 32767) + 0x8000;
                        low = min(low, sample);
                        high = max(high, sample);
                    }
                    const unsigned int ylow =
                        ybase - (low * channel_height) / 65536;
                    const unsigned int yhigh =
                        ybase - (high * channel_height) / 65536;
                    gdk_gc_set_foreground(gc, wave_get_channel_color(canvas->wave, ch));
                    gdk_draw_line(drawable, gc, x, ylow, x, yhigh);
                }
            }
        }
        free(data);

        /* Also draw the current selection if one exists and it overlaps
         * the displayed region. */
        if (wave_select_len(canvas->wave) > 0) {
            const int64_t selection_start =
                (int64_t)(wave_position(canvas->wave) - display_start)
                / (int64_t)zoom;
            const int64_t selection_end =
                (int64_t)(wave_position(canvas->wave)
                          + wave_select_len(canvas->wave) - 1 - display_start)
                / (int64_t)zoom;
            if (selection_start < width && selection_end >= 0) {
                const unsigned int left =
                    (unsigned int)lbound(selection_start, 0);
                const unsigned int right =
                    (unsigned int)ubound(selection_end, width);
                static const GdkColor green_xor = {.pixel = 0x0000FF00};
                gdk_gc_set_foreground(gc, &green_xor);
                gdk_gc_set_function(gc, GDK_XOR);
                gdk_draw_rectangle(drawable, gc, TRUE,
                                   left, 0, right - left + 1, height);
            }
        }

    }  // if (data)

    /* Destroy the graphics context we created. */
    gdk_gc_destroy(gc);

    return TRUE;
}

/*----------------------------------*/

static gboolean cb_event_focus_in(GtkWidget *widget, GdkEvent *event,
                                  CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    /* If we get the focus, switch the active wave to ourselves unless
     * another wave file is currently being played. */
    if (!ui_in_playback() && manager_get_active_wave() != canvas->wave) {
        gtk_widget_grab_focus(widget);
        manager_set_active_wave(canvas->wave);
        ui_update();
    }

    /* We return TRUE here to prevent GTK+ from generating an expose event
     * for the widget and causing an unnecessary redraw.
     * (http://trac.wxwidgets.org/ticket/3056 suggests that this expose
     * event generation is used by GTK+ to draw focus indicators, but since
     * we rely on the window manager's focus management, we don't need to
     * draw anything ourselves.) */
    return TRUE;
}

/*----------------------------------*/

static gboolean cb_event_key_press(GtkWidget *widget, GdkEventKey *event,
                                   CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return FALSE);

    /* Don't act on any keypresses if a dialog window is displayed.  (GTK+
     * will route them to the dialog itself.) */
    if (ui_in_dialog()) {
        return FALSE;
    }

    const unsigned int state = event->state & (GDK_SHIFT_MASK
                                             | GDK_CONTROL_MASK
                                             | GDK_MOD1_MASK);  // i.e. Alt

    switch (event->keyval) {

      case GDK_Return:
      case GDK_KP_Enter:
        /* Update the canvas window position based on the current cursor
         * position or selected range, like a middle click. */
        if (!ui_in_playback()) {
            canvas_refresh(canvas, 1);
        }
        return TRUE;

      case GDK_Left:
      case GDK_Right: {
        /* Treat like a scroll left or right. */
        GdkEventScroll scroll_event;
        scroll_event.type = GDK_SCROLL;
        scroll_event.window = event->window;
        scroll_event.send_event = 1;  // This is a fake event
        scroll_event.time = event->time;
        scroll_event.x = scroll_event.y = 0;  // Not used, so this is okay
        scroll_event.state = event->state;
        scroll_event.direction = (event->keyval==GDK_Left
                                  ? GDK_SCROLL_LEFT : GDK_SCROLL_RIGHT);
        scroll_event.device = NULL;  // Not used
        scroll_event.x_root = scroll_event.y_root = 0;  // Not used
        cb_event_scroll(widget, &scroll_event, canvas);
        return TRUE;
      }  // case GDK_Left, GDK_Right

      case GDK_Home:
        if (state == 0) {
            /* Move the cursor position to the left edge of the canvas
             * window. */
            if (wave_position(canvas->wave) != canvas->display_start) {
                const uint64_t limit =
                    wave_samples(canvas->wave) - wave_select_len(canvas->wave);
                wave_set_position(canvas->wave,
                                  ubound(canvas->display_start, limit), 0);
                ui_update();
            }
            return TRUE;
        } else if (state == GDK_SHIFT_MASK) {
            /* Move the left edge of the selection to the left edge of the
             * canvas window.  (If the selection is entirely to the left of
             * the window, we instead clear it and reposition the cursor at
             * the left edge of the window.  This should be impossible,
             * since either the cursor position or the selection is always
             * kept visible, but we check anyway to avoid unsigned integer
             * wraparound during subtraction.) */
            const uint64_t select_end =
                wave_position(canvas->wave) + wave_select_len(canvas->wave);
            if (wave_position(canvas->wave) != canvas->display_start) {
                if (select_end > canvas->display_start) {
                    wave_set_selection(canvas->wave, canvas->display_start,
                                       select_end - canvas->display_start, 0);
                } else {
                    /* The selection is entirely to the left of the window. */
                    wave_set_selection(canvas->wave, canvas->display_start,
                                       0, 0);
                }
                ui_update();
            }
            return TRUE;
        }
        break;

      case GDK_End:
        if (state == 0) {
            /* Move the cursor position to the right edge of the canvas
             * window. */
            const uint64_t right_edge = canvas->display_start
                + wave_zoom(canvas->wave) * (canvas_width(canvas) - 1);
            if (wave_position(canvas->wave) != right_edge) {
                const uint64_t limit =
                    wave_samples(canvas->wave) - wave_select_len(canvas->wave);
                wave_set_position(canvas->wave, ubound(right_edge, limit), 0);
                ui_update();
            }
            return TRUE;
        } else if (state == GDK_SHIFT_MASK) {
            /* Move the right edge of the selection to the right edge of
             * the canvas window.  Note that we calculate "right_edge"
             * differently than above, since we want to include the last
             * pixel in the selection. */
            const uint64_t right_edge = canvas->display_start
                + wave_zoom(canvas->wave) * canvas_width(canvas);
            const uint64_t select_end =
                wave_position(canvas->wave) + wave_select_len(canvas->wave);
            if (select_end != right_edge) {
                if (wave_position(canvas->wave) < right_edge) {
                    wave_set_select_len(canvas->wave,
                                        right_edge - wave_position(canvas->wave), 0);
                } else {
                    wave_set_selection(canvas->wave, right_edge, 0, 0);
                }
                ui_update();
            }
            return TRUE;
        }
        break;

    }  // switch (event->keyval)

    /* If it's not a key we handle on our own, pass it to the main window. */
    return gtk_widget_event(ui_get_widget("main_window"), (GdkEvent *)event);
}

/*----------------------------------*/

static gboolean cb_event_button_press(GtkWidget *widget, GdkEventButton *event,
                                      CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    /* Don't react to events while playing audio. */
    if (ui_in_playback()) {
        return FALSE;
    }

    /* On a middle-click (button 2), reposition the cursor at the left
     * edge of the canvas. */
    if (event->button == 2) {
        canvas_refresh(canvas, 1);
        return TRUE;
    }

    /* Otherwise, only react to the left button (button 1). */
    if (event->button != 1) {
        return FALSE;
    }

    /* Set the position to the clicked location, and cancel any selection. */
    GtkAllocation allocation;
    gtk_widget_get_allocation(widget, &allocation);
    const unsigned int width = allocation.width;
    const unsigned int x = (unsigned int)bound(event->x, 0, width-1);
    const uint64_t new_position =
        ubound(canvas->display_start + (x * wave_zoom(canvas->wave)),
               wave_samples(canvas->wave));
    wave_set_selection(canvas->wave, new_position, 0, 0);
    ui_update();

    /* Save the clicked position so dragging works properly. */
    canvas->clicked_position = wave_position(canvas->wave);

    return TRUE;
}

/*----------------------------------*/

static gboolean cb_event_motion(GtkWidget *widget, GdkEventMotion *event,
                                CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    /* Don't react to events while playing audio. */
    if (ui_in_playback()) {
        return FALSE;
    }

    /* Only react to the left button (button 1). */
    if (!(event->state & GDK_BUTTON1_MASK)) {
        return FALSE;
    }

    /* Adjust the size of the selection based on the current mouse position. */
    GtkAllocation allocation;
    gtk_widget_get_allocation(widget, &allocation);
    const unsigned int width = allocation.width;
    const unsigned int x = (unsigned int)bound(event->x, 0, width);
    const uint64_t new_position =
        ubound(canvas->display_start + (x * wave_zoom(canvas->wave)),
               wave_samples(canvas->wave));
    if (new_position < canvas->clicked_position) {
        wave_set_selection(canvas->wave, new_position,
                           canvas->clicked_position - new_position, 0);
    } else {
        wave_set_selection(canvas->wave, canvas->clicked_position,
                           new_position - canvas->clicked_position, 0);
    }
    ui_update();

    return TRUE;
}

/*----------------------------------*/

static gboolean cb_event_scroll(GtkWidget *widget, GdkEventScroll *event,
                                CanvasWindow *canvas)
{
    PRECOND(canvas != NULL, return TRUE);

    const unsigned int state = event->state & (GDK_SHIFT_MASK
                                             | GDK_CONTROL_MASK
                                             | GDK_MOD1_MASK);  // i.e. Alt

    /* Don't react to events while playing audio. */
    if (ui_in_playback()) {
        return FALSE;
    }

    switch (event->direction) {
      case GDK_SCROLL_UP: {
        /* On a scroll up, increment the zoom level by one (unshifted) or
         * multiply it by 10 (with Ctrl). */
        const uint64_t limit =
            lbound(wave_samples(canvas->wave) / canvas_width(canvas), 1);
        if (wave_zoom(canvas->wave) < limit) {
            if (state == 0) {
                wave_set_zoom(canvas->wave, wave_zoom(canvas->wave) + 1);
                ui_update();
            } else if (state == GDK_CONTROL_MASK) {
                if (wave_zoom(canvas->wave) > limit/10) {
                    wave_set_zoom(canvas->wave, limit);
                } else {
                    wave_set_zoom(canvas->wave, wave_zoom(canvas->wave) * 10);
                }
                ui_update();
            }
        }
        break;
      }  // case GDK_SCROLL_UP

      case GDK_SCROLL_DOWN:
        /* On a scroll down, decrement the zoom level by one (unshifted) or
         * divide it by 10 (with Ctrl). */
        if (wave_zoom(canvas->wave) > 1) {
            if (state == 0) {
                wave_set_zoom(canvas->wave, wave_zoom(canvas->wave) - 1);
                ui_update();
            } else if (state == GDK_CONTROL_MASK) {
                if (wave_zoom(canvas->wave) < 10) {
                    wave_set_zoom(canvas->wave, 1);
                } else {
                    wave_set_zoom(canvas->wave, wave_zoom(canvas->wave) / 10);
                }
                ui_update();
            }
        }
        break;

      case GDK_SCROLL_LEFT:
        if ((state & ~GDK_CONTROL_MASK) == 0) {
            /* Move the cursor position left one pixel (100 pixels if
             * Ctrl is pressed), also scrolling the canvas display if
             * appropriate. */
            const uint64_t position = wave_position(canvas->wave);
            const uint64_t zoom = wave_zoom(canvas->wave);
            const uint64_t distance = (state & GDK_CONTROL_MASK) ? 100 : 1;
            if (position > 0) {
                const uint64_t move = min(position, zoom * distance);
                const uint64_t new_position = position - move;
                wave_set_position(canvas->wave, new_position, 0);
                if (new_position < canvas->display_start) {
                    /* Rather than just shifting to the new position, which
                     * will cause the appearance of the waveform to change
                     * if we shift a fractional number of pixels, we alter
                     * the display position a pixel at a time until the
                     * cursor position is visible. */
                    while (new_position < canvas->display_start) {
                        const uint64_t scroll =
                            min(canvas->display_start, zoom);
                        canvas->display_start -= scroll;
                    }
                    canvas_refresh(canvas, 0);
                }
                ui_update();
            }

        } else if ((state & ~GDK_CONTROL_MASK) == GDK_SHIFT_MASK) {
            /* Extend the current selection left one pixel (100 pixels
             * if Ctrl is pressed). */
            const uint64_t position = wave_position(canvas->wave);
            const uint64_t select_len = wave_select_len(canvas->wave);
            const uint64_t zoom = wave_zoom(canvas->wave);
            const uint64_t distance = (state & GDK_CONTROL_MASK) ? 100 : 1;
            if (position > 0) {
                const uint64_t extend = min(position, zoom * distance);
                wave_set_selection(canvas->wave, position - extend,
                                   select_len + extend, 0);
                ui_update();
            }
        }
        break;

      case GDK_SCROLL_RIGHT:
        if ((state & ~GDK_CONTROL_MASK) == 0) {
            /* Move the cursor position left one pixel (100 pixels if
             * Ctrl is pressed), also scrolling the canvas display if
             * appropriate. */
            const uint64_t position = wave_position(canvas->wave);
            const uint64_t zoom = wave_zoom(canvas->wave);
            const uint64_t distance = (state & GDK_CONTROL_MASK) ? 100 : 1;
            const uint64_t limit =
                wave_samples(canvas->wave) - wave_select_len(canvas->wave);
            if (position < limit) {
                const uint64_t move = min(limit - position, zoom * distance);
                const uint64_t new_position = position + move;
                wave_set_position(canvas->wave, new_position, 0);
                const uint64_t display_samples = canvas_width(canvas) * zoom;
                if (new_position >= canvas->display_start + display_samples) {
                    uint64_t display_limit;
                    if (wave_samples(canvas->wave) < display_samples/2) {
                        display_limit = 0;
                    } else {
                        display_limit =
                            wave_samples(canvas->wave) - display_samples/2;
                    }
                    while (new_position >= canvas->display_start + display_samples) {
                        const uint64_t scroll =
                            min(display_limit - canvas->display_start, zoom);
                        canvas->display_start += scroll;
                    }
                    canvas_refresh(canvas, 0);
                }
                ui_update();
            }

        } else if ((state & ~GDK_CONTROL_MASK) == GDK_SHIFT_MASK) {
            /* Extend the current selection right one pixel (100 pixels
             * if Ctrl is pressed). */
            const uint64_t select_len = wave_select_len(canvas->wave);
            const uint64_t zoom = wave_zoom(canvas->wave);
            const uint64_t distance = (state & GDK_CONTROL_MASK) ? 100 : 1;
            const uint64_t limit =
                wave_samples(canvas->wave) - wave_position(canvas->wave);
            if (select_len < limit) {
                const uint64_t extend =
                    min(limit - select_len, zoom * distance);
                wave_set_select_len(canvas->wave, select_len + extend, 0);
                ui_update();
            }
        }
        break;

    }  // switch (event->direction)

    return TRUE;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
