/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/ui.c: Main user interface management.  Canvas windows for individual
 * wave files are _not_ handled here; they are instead handled by canvas.c.
 */

#include "common.h"
#include "audio.h"
#include "manager.h"
#include "ui.h"
#include "wave.h"

#include <assert.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/* User interface XML definition (generated from xwaven.glade) */
#include "ui-glade.i"

/* User interface object */
static GtkBuilder *ui;

/* Last clicked position (used by dragging code) */
static uint64_t clicked_position;

/*----------------------------------*/

/* Clipboard data (NULL indicates that the clipboard is empty) */
static float *clipboard;

/* Length of clipboard data in samples (only valid if clipboard != NULL) */
static uint64_t clipboard_length;

/* Parameters of clipboard data (only used for Paste as New File) */
static unsigned int clipboard_format;
static unsigned int clipboard_channels;
static unsigned int clipboard_srate;

/*----------------------------------*/

/* Flag: Are we currently playing audio? */
static int in_playback = 0;

/* Playback parameters (only valid during playback) */
static WaveFile *playback_wave; // Wave file to play
static uint64_t playback_start; // First sample to play
static uint64_t playback_end;   // Last sample to play + 1
static uint64_t playback_pos;   // Current playback position
static guint playback_idle_id;  // GTK+ idle function ID

/* Playback position counter, in pixels from the left edge of the main
 * canvas; initialize to -1, then the playback routine will update this
 * after each iteration. */
static int playback_x;

/*----------------------------------*/

/* Flag: Are we currently recording (or monitoring) audio? */
static int in_record = 0;

/* Flag: Are we actually recording rather than just monitoring? */
static int recording = 0;

/* Recording parameters (only valid during playback) */
static WaveFile *record_wave;   // Wave file to receive recorded data
static float *record_buffer;    // Buffer into which to record data
static uint64_t record_bufsize; // Size of buffer in samples
static uint64_t record_length;  // Number of samples recorded
static unsigned int record_overruns; // Number of overruns detected
static int record_clipped;      // Flag: Has possible clipping been detected?
static guint record_idle_id;    // GTK+ idle function ID

/* Current high and low levels on each channel (initialize to zero before
 * starting record_audio()) */
static float record_low_l, record_high_l;
static float record_low_r, record_high_r;

/* Most recently active wave file (will be reactivated if recording is
 * cancelled) */
static WaveFile *record_last_wave;

/*----------------------------------*/

// FIXME: helper flags for ui_in_dialog(), necessitated by the lack of
// working Glade support for message windows
static int in_show_error_dialog;
static int in_close_wave_dialog;

/*************************************************************************/

/* Widget callback declarations.  These have to be extern so GtkBuilder can
 * look up the symbols at UI creation time. */

extern gboolean uicb_event_button_press(GtkWidget *widget, GdkEventButton *event);
extern gboolean uicb_event_motion(GtkWidget *widget, GdkEventMotion *event);
extern gboolean uicb_event_delete(GtkWidget *widget);
extern gboolean uicb_event_expose(GtkWidget *widget, GdkEventExpose *event);
extern gboolean uicb_event_key_press(GtkWidget *widget, GdkEventKey *event);

extern void uicb_update_position(GtkWidget *widget);
extern void uicb_update_select_len(GtkWidget *widget);
extern void uicb_update_zoom(GtkWidget *widget);

extern void uicb_action_new(GtkWidget *widget);
extern void uicb_action_open(GtkWidget *widget);
extern void uicb_action_save(GtkWidget *widget);
extern void uicb_action_save_as(GtkWidget *widget);
extern void uicb_action_close(GtkWidget *widget);
extern void uicb_action_quit(GtkWidget *widget);
extern void uicb_action_cut(GtkWidget *widget);
extern void uicb_action_copy(GtkWidget *widget);
extern void uicb_action_paste(GtkWidget *widget);
extern void uicb_action_paste_new(GtkWidget *widget);
extern void uicb_action_delete(GtkWidget *widget);
extern void uicb_action_stop(GtkWidget *widget);
extern void uicb_action_play(GtkWidget *widget);
extern void uicb_action_record(GtkWidget *widget);
extern void uicb_action_about(GtkWidget *widget);

extern void uicb_recdialog_cancel(GtkWidget *widget);
extern void uicb_recdialog_action(GtkWidget *widget, gint response_id);
extern void uicb_recdialog_stop(GtkWidget *widget);
extern gboolean uicb_recdialog_monitor_expose(GtkWidget *widget,
                                              GdkEventExpose *event);

/* Other local function declarations */

static int get_properties(const char *prompt, unsigned int *bits_ptr,
                          unsigned int *channels_ptr, unsigned int *srate_ptr);
static gboolean play_audio(void);
static gboolean record_audio(void);

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

/**
 * ui_init:  Set up the GTK+ user interface.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     Nonzero on success, zero on failure
 */
int ui_init(void)
{
    /* Create the interface. */
    ui = gtk_builder_new();
    assert(gtk_builder_add_from_string(ui, ui_glade, sizeof(ui_glade)-1, NULL));
    gtk_builder_connect_signals(ui, NULL);

    /* Set parameters libglade doesn't support. */
#if 0  // FIXME: message dialogs are currently broken in libglade (see xwaven-msgdialog.glade)
    GtkWidget *image =
        gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_DIALOG);
    gtk_message_dialog_set_image(
        GTK_MESSAGE_DIALOG(ui_get_widget("confirm_close_dialog")), image
    );
    gtk_widget_show(image);
#endif

    /* Create a file filter (since libglade can't deal with those itself)
     * and attach it to the file chooser dialogs. */
    GtkFileFilter *filter = gtk_file_filter_new();
    if (!filter) {
        DMSG("Failed to create file filter");
        return 0;
    }
    gtk_file_filter_add_pattern(filter, "*.wav");
    gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(ui_get_widget("open_dialog")), filter);
    gtk_file_chooser_set_filter(GTK_FILE_CHOOSER(ui_get_widget("save_as_dialog")), filter);

    /* Position the main program window in the upper-left corner of the
     * screen, for consistency. */
    gtk_window_move(GTK_WINDOW(ui_get_widget("main_window")), 0, 0);

    /* Display the main program window. */
    gtk_widget_show(ui_get_widget("main_window"));

    /* Update the status line with default values before returning. */
    ui_update();

    return 1;
}

/*************************************************************************/

/**
 * ui_get_widget:  Return a pointer to the named widget in the user
 * interface.  A shortcut for "gtk_builder_get_object(ui,name)".
 *
 * [Parameters]
 *     name: Widget name
 * [Return value]
 *     Pointer to widget, or NULL if not found
 */
GtkWidget *ui_get_widget(const char *name)
{
    PRECOND(ui != NULL, return NULL);
    PRECOND(name != NULL, return NULL);
    return GTK_WIDGET(gtk_builder_get_object(ui, name));
}

/*************************************************************************/

/**
 * ui_update:  Update the UI after a change in state.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     None
 */
void ui_update(void)
{
    /* Block signal handlers during the update so they don't generate
     * additional update events. */
    g_signal_handlers_block_by_func(ui_get_widget("position"),
                                    uicb_update_position, NULL);
    g_signal_handlers_block_by_func(ui_get_widget("select_len"),
                                    uicb_update_select_len, NULL);
    g_signal_handlers_block_by_func(ui_get_widget("zoom"),
                                    uicb_update_zoom, NULL);

    /* Update various UI elements. */

    #define SET_ENTRY_FIELD(name,enabled,value,low,high,step)  do {         \
        GtkSpinButton *field = GTK_SPIN_BUTTON(ui_get_widget(#name));       \
        GtkAdjustment *adjustment = gtk_spin_button_get_adjustment(field);  \
        gtk_adjustment_set_lower(adjustment, (low));                        \
        gtk_adjustment_set_upper(adjustment, (high));                       \
        gtk_adjustment_set_step_increment(adjustment, (step));              \
        gtk_adjustment_set_page_increment(adjustment, (step)*10);           \
        gtk_spin_button_set_value(field, (value));                          \
        gtk_widget_set_sensitive(GTK_WIDGET(field), (enabled));             \
    } while (0)
    #define SET_STATUS(name,string) \
        gtk_label_set_text(GTK_LABEL(ui_get_widget("status_" #name)), (string))
    #define SET_STATUSF(name,format,...)  do {                              \
        char buf[100];                                                      \
        snprintf(buf, sizeof(buf), format, __VA_ARGS__);                    \
        gtk_label_set_text(GTK_LABEL(ui_get_widget("status_" #name)), buf); \
    } while (0)
    #define SET_SENSITIVE(name,flag) \
        gtk_widget_set_sensitive(ui_get_widget(#name), (flag));

    WaveFile *wave = manager_get_active_wave();

    if (wave) {

        const char *path = wave_path(wave);
        const uint64_t samples = wave_samples(wave);
        const uint64_t bytes =
            samples * (wave_bits(wave)/8) * wave_channels(wave)
            + 44;  // WAV file header size
        const double secs = (double)samples / (double)wave_srate(wave);

        const uint64_t max_zoom = lbound(samples / wave_canvas_width(wave), 1);
        if (wave_zoom(wave) > max_zoom) {
            wave_set_zoom(wave, max_zoom);
        }

        char titlebuf[1000], channelbuf[10];
        snprintf(titlebuf, sizeof(titlebuf), "xwaven: %s",
                 path ? path : "<New file>");
        snprintf(channelbuf, sizeof(channelbuf), "%dch", wave_channels(wave));
        gtk_window_set_title(GTK_WINDOW(ui_get_widget("main_window")),
                             titlebuf);
        SET_ENTRY_FIELD(position,   !in_playback,
                        in_playback ? playback_pos : wave_position(wave),
                        0,
                        in_playback ? samples : samples - wave_select_len(wave),
                        wave_zoom(wave));
        SET_ENTRY_FIELD(select_len, !in_playback, wave_select_len(wave),
                        0, samples - wave_zoom(wave), wave_zoom(wave));
        SET_ENTRY_FIELD(zoom,       !in_playback, wave_zoom(wave),
                        1, max_zoom, 1);
        SET_STATUS (pathname, path ? path : "<New file>");
        SET_STATUS (writable, wave_writable(wave) ? "RW" : "RO");
        SET_STATUS (modified, wave_modified(wave) ? "m" : "-");
        SET_STATUS (channels, wave_channels(wave)==1 ? "Mono" :
                              wave_channels(wave)==2 ? "Stereo" :
                              wave_channels(wave)==6 ? "Surround" :
                              channelbuf);
        SET_STATUSF(bits,     "%u bit", wave_bits(wave));
        SET_STATUSF(srate,    "%u Hz", wave_srate(wave));
        SET_STATUSF(filesize, "%llu kB", (long long)(bytes / 1000));
        SET_STATUSF(duration, "%.2f s", secs);

        const gboolean writable = wave_path(wave)!=NULL && wave_writable(wave);
        const gboolean modified = wave_modified(wave);
        const gboolean have_sel = (wave_select_len(wave) > 0);
        const gboolean have_clip = (clipboard != NULL);
        SET_SENSITIVE(menuitem_New,      !in_playback);
        SET_SENSITIVE(menuitem_Open,     !in_playback);
        SET_SENSITIVE(menuitem_Save,     !in_playback && writable && modified);
        SET_SENSITIVE(menuitem_Save_As,  !in_playback);
        SET_SENSITIVE(menuitem_Close,    !in_playback);
        SET_SENSITIVE(menuitem_Cut,      !in_playback && have_sel);
        SET_SENSITIVE(menuitem_Copy,     !in_playback && have_sel);
        SET_SENSITIVE(menuitem_Paste,    !in_playback && have_clip);
        SET_SENSITIVE(menuitem_PasteNew, !in_playback && have_clip);
        SET_SENSITIVE(menuitem_Delete,   !in_playback && have_sel);
        SET_SENSITIVE(button_New,        !in_playback);
        SET_SENSITIVE(button_Open,       !in_playback);
        SET_SENSITIVE(button_Save,       !in_playback && writable && modified);
        SET_SENSITIVE(button_Cut,        !in_playback && have_sel);
        SET_SENSITIVE(button_Copy,       !in_playback && have_sel);
        SET_SENSITIVE(button_Paste,      !in_playback && have_clip);
        SET_SENSITIVE(button_Delete,     !in_playback && have_sel);
        SET_SENSITIVE(button_Stop,        in_playback);
        SET_SENSITIVE(button_Play,       !in_playback);
        SET_SENSITIVE(button_Record,     !in_playback);

    } else {  // !wave

        gtk_window_set_title(GTK_WINDOW(ui_get_widget("main_window")),
                             "xwaven");
        SET_ENTRY_FIELD(position,   FALSE, 0, 0, 0, 0);
        SET_ENTRY_FIELD(select_len, FALSE, 0, 0, 0, 0);
        SET_ENTRY_FIELD(zoom,       FALSE, 1, 1, 1, 0);
        SET_STATUS (pathname, "<No file>");
        SET_STATUS (writable, "");
        SET_STATUS (modified, "");
        SET_STATUS (channels, "");
        SET_STATUS (bits,     "");
        SET_STATUS (srate,    "");
        SET_STATUS (filesize, "");
        SET_STATUS (duration, "");

        const gboolean have_clip = (clipboard != NULL);
        SET_SENSITIVE(menuitem_New,      TRUE);
        SET_SENSITIVE(menuitem_Open,     TRUE);
        SET_SENSITIVE(menuitem_Save,     FALSE);
        SET_SENSITIVE(menuitem_Save_As,  FALSE);
        SET_SENSITIVE(menuitem_Close,    FALSE);
        SET_SENSITIVE(menuitem_Cut,      FALSE);
        SET_SENSITIVE(menuitem_Copy,     FALSE);
        SET_SENSITIVE(menuitem_Paste,    FALSE);
        SET_SENSITIVE(menuitem_PasteNew, have_clip);
        SET_SENSITIVE(menuitem_Delete,   FALSE);
        SET_SENSITIVE(button_New,        TRUE);
        SET_SENSITIVE(button_Open,       TRUE);
        SET_SENSITIVE(button_Save,       FALSE);
        SET_SENSITIVE(button_Cut,        FALSE);
        SET_SENSITIVE(button_Copy,       FALSE);
        SET_SENSITIVE(button_Paste,      FALSE);
        SET_SENSITIVE(button_Delete,     FALSE);
        SET_SENSITIVE(button_Stop,       FALSE);
        SET_SENSITIVE(button_Play,       FALSE);

    }

    #undef SET_ENTRY_FIELD
    #undef SET_STATUS
    #undef SET_STATUSF
    #undef SET_SENSITIVE

    /* Trigger a redraw of the main canvas area. */
    GtkWidget *main_canvas = ui_get_widget("main_canvas");
    GtkAllocation allocation;
    gtk_widget_get_allocation(main_canvas, &allocation);
    allocation.x = allocation.y = 0;
    gdk_window_invalidate_rect(gtk_widget_get_window(main_canvas),
                               &allocation, FALSE);

    /* Unblock the signal handlers we blocked above. */
    g_signal_handlers_unblock_by_func(ui_get_widget("position"),
                                      uicb_update_position, NULL);
    g_signal_handlers_unblock_by_func(ui_get_widget("select_len"),
                                      uicb_update_select_len, NULL);
    g_signal_handlers_unblock_by_func(ui_get_widget("zoom"),
                                      uicb_update_zoom, NULL);
}

/*************************************************************************/

/**
 * ui_show_error:  Display an error dialog and wait for the user to
 * dismiss it.
 *
 * [Parameters]
 *     format: Error message format string (printf()-style)
 *        ...: Format arguments
 * [Return value]
 *     None
 */
void ui_show_error(const char *format, ...)
{
    char buf[10000];
    va_list args;
    va_start(args, format);
    vsnprintf(buf, sizeof(buf), format, args);
    va_end(args);
    GtkWidget *dialog = gtk_message_dialog_new(
        GTK_WINDOW(ui_get_widget("main_window")),
        GTK_DIALOG_DESTROY_WITH_PARENT,
        GTK_MESSAGE_ERROR,
        GTK_BUTTONS_OK,
        "%s", buf
    );
    gtk_window_set_title(GTK_WINDOW(dialog), "Error");
    in_show_error_dialog = 1;
    gtk_dialog_run(GTK_DIALOG(dialog));
    in_show_error_dialog = 0;
    gtk_widget_destroy(dialog);
}

/*************************************************************************/

/**
 * ui_close_wave:  Close the given wave file, first prompting the user to
 * save changes if necessary.
 *
 * [Parameters]
 *     wave: Wave file to close
 * [Return value]
 *     Nonzero if the wave file was closed, zero otherwise (e.g., if the
 *     user cancelled the confirmation dialog)
 */
int ui_close_wave(WaveFile *wave)
{
    PRECOND(wave != NULL, return 0);

    if (wave_modified(wave)) {
        const char *path = wave_path(wave);
        char buf[1000];
        snprintf(buf, sizeof(buf),
                 path ? "Save changes to the file \"%s\" before closing?"
                      : "Save changes to this new file before closing?",
                 path);
#if 0  // FIXME: message dialogs are currently broken in libglade (see xwaven-msgdialog.glade)
        GtkWidget *dialog = ui_get_widget("confirm_close_dialog");
        gtk_message_dialog_set_markup(GTK_MESSAGE_DIALOG(dialog), buf);
        gtk_message_dialog_format_secondary_text(
            GTK_MESSAGE_DIALOG(dialog),
            path ? "If you don't save the file, your changes will be lost."
                 : "If you don't save the file, its contents will be lost."
        );
        const int result = gtk_dialog_run(GTK_DIALOG(dialog));
        gtk_widget_hide(dialog);
#else
        GtkWidget *dialog = gtk_message_dialog_new(
            GTK_WINDOW(ui_get_widget("main_window")),
            GTK_DIALOG_DESTROY_WITH_PARENT,
            GTK_MESSAGE_OTHER,
            GTK_BUTTONS_NONE,
            "%s", buf
        );
        gtk_window_set_title(GTK_WINDOW(dialog), "Confirm Close");
        GtkWidget *image =
            gtk_image_new_from_stock(GTK_STOCK_SAVE, GTK_ICON_SIZE_DIALOG);
        gtk_message_dialog_set_image(GTK_MESSAGE_DIALOG(dialog), image);
        gtk_widget_show(image);
        gtk_message_dialog_format_secondary_text(
            GTK_MESSAGE_DIALOG(dialog),
            path ? "If you don't save the file, your changes will be lost."
                 : "If you don't save the file, its contents will be lost."
        );
        GtkWidget *button = gtk_button_new_with_label("Don't Save");
        gtk_button_set_image(
            GTK_BUTTON(button),
            gtk_image_new_from_stock(GTK_STOCK_DELETE, GTK_ICON_SIZE_BUTTON)
        );
        gtk_dialog_add_action_widget(GTK_DIALOG(dialog),
                                     button, GTK_RESPONSE_CLOSE);
        gtk_widget_show(button);
        gtk_dialog_add_buttons(GTK_DIALOG(dialog),
                               GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                               GTK_STOCK_SAVE, GTK_RESPONSE_OK,
                               NULL);
        gtk_dialog_set_default_response(GTK_DIALOG(dialog),
                                        GTK_RESPONSE_CANCEL);
        in_close_wave_dialog = 1;
        const int result = gtk_dialog_run(GTK_DIALOG(dialog));
        in_close_wave_dialog = 0;
        gtk_widget_destroy(dialog);
#endif
        if (result == GTK_RESPONSE_OK) {
            if (path && wave_writable(wave)) {
                uicb_action_save(NULL);
            } else {
                uicb_action_save_as(NULL);
            }
            if (wave_modified(wave)) {
                DMSG("Save failed");
                return 0;
            }
        } else if (result != GTK_RESPONSE_CLOSE) {
            return 0;  // Cancelled by user
        }
    }

    manager_delete_wave(wave);
    wave_destroy(wave);
    return 1;
}

/*************************************************************************/

/**
 * ui_in_dialog:  Return whether a dialog window is currently displayed.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     True (nonzero) if a dialog is currently displayed, else false (zero)
 */
int ui_in_dialog(void)
{
    return gtk_widget_get_visible(ui_get_widget("about_dialog"))
        || gtk_widget_get_visible(ui_get_widget("open_dialog"))
        || gtk_widget_get_visible(ui_get_widget("properties_dialog"))
        || gtk_widget_get_visible(ui_get_widget("record_dialog"))
        || gtk_widget_get_visible(ui_get_widget("save_as_dialog"))
        || in_show_error_dialog
        || in_close_wave_dialog
        ;
}

/*----------------------------------*/

/**
 * ui_in_playback:  Return whether a wave file is currently playing.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     True (nonzero) if a wave file is currently playing, else false (zero)
 */
int ui_in_playback(void)
{
    return in_playback;
}

/*************************************************************************/
/*************************** Widget callbacks ****************************/
/*************************************************************************/

/**
 * uicb_event_*:  Callbacks for GDK events.
 *
 * [Parameters]
 *     widget: Relevant widget
 *      event: Event data
 * [Return value]
 *     TRUE if further GTK+ processing should be halted, else FALSE
 */

gboolean uicb_event_delete(GtkWidget *widget)
{
    uicb_action_quit(widget);
    return TRUE;  // Don't destroy the window immediately.
}

/*----------------------------------*/

gboolean uicb_event_expose(GtkWidget *widget, GdkEventExpose *event)
{
    /* Create a graphics context for drawing. */
    GdkDrawable *drawable = GDK_DRAWABLE(gtk_widget_get_window(widget));
    GdkGC *gc = gdk_gc_new(drawable);
    if (!gc) {
        DMSG("Failed to create GC");
        return FALSE;
    }

    /* Retrieve the size of the drawing area. */
    int width, height;
    gdk_drawable_get_size(drawable, &width, &height);

    /* Start out by painting the background white.  To save time, we only
     * paint the portion of the canvas covered by the exposure area (but we
     * paint the entire vertical region to simplify wave-drawing logic). */
    static const GdkColor white = {.pixel = 0xFFFFFFFF};
    gdk_gc_set_foreground(gc, &white);
    gdk_draw_rectangle(drawable, gc, TRUE,
                       event->area.x, 0, event->area.width, height);

    /* If we have an active, non-empty wave, draw its data. */
    WaveFile *wave = manager_get_active_wave();
    if (wave && wave_samples(wave) > 0) {
        const uint16_t *canvas_data = wave_get_full_canvas(wave, width);
        if (canvas_data) {
            const unsigned int channels = wave_channels(wave);
            const unsigned int channel_height = height / channels;
            canvas_data += event->area.x * (2*channels);
            unsigned int x;
            for (x = event->area.x; x < event->area.x+event->area.width; x++) {
                unsigned int ch;
                for (ch = 0; ch < channels; ch++, canvas_data += 2) {
                    const unsigned int ybase = (ch+1) * channel_height - 1;
                    const unsigned int low   = canvas_data[0];
                    const unsigned int high  = canvas_data[1];
                    const unsigned int ylow  =
                        ybase - (low * channel_height) / 65536;
                    const unsigned int yhigh =
                        ybase - (high * channel_height) / 65536;
                    gdk_gc_set_foreground(gc, wave_get_channel_color(wave, ch));
                    gdk_draw_line(drawable, gc, x, ylow, x, yhigh);
                }
            }
        }
        /* Draw the current playback position if we're playing this file. */
        if (in_playback && playback_wave == wave && playback_x >= 0) {
            static const GdkColor black = {.pixel = 0xFF000000};
            gdk_gc_set_foreground(gc, &black);
            gdk_draw_line(drawable, gc, playback_x, 0, playback_x, height-1);
        }
        /* Also draw the current cursor position or selection. */
        const unsigned int position_start =
            ubound(wave_position(wave) * width / wave_samples(wave), width-1);
        const unsigned int position_end =
            ubound((wave_position(wave) + wave_select_len(wave) - 1)
                   * width / wave_samples(wave), width-1);
        if (wave_select_len(wave) > 0) {
            static const GdkColor green_xor = {.pixel = 0x0000FF00};
            gdk_gc_set_foreground(gc, &green_xor);
            gdk_gc_set_function(gc, GDK_XOR);
            gdk_draw_rectangle(drawable, gc, TRUE, position_start, 0,
                               position_end - position_start + 1, height);
        } else {
            static const GdkColor black = {.pixel = 0xFF000000};
            gdk_gc_set_foreground(gc, &black);
            gdk_draw_line(drawable, gc, position_start, 0,
                          position_start, height-1);
        }
    }  // if (wave)

    /* Destroy the graphics context we created. */
    gdk_gc_destroy(gc);

    return TRUE;
}

/*----------------------------------*/

gboolean uicb_event_key_press(GtkWidget *widget, GdkEventKey *event)
{
    WaveFile *wave = manager_get_active_wave();
    const unsigned int state = event->state & (GDK_SHIFT_MASK
                                             | GDK_CONTROL_MASK
                                             | GDK_MOD1_MASK);  // i.e. Alt

    switch (event->keyval) {

      case ' ':
        if (state == 0) {
            /* Start or stop playback. */
            if (!wave) {
                break;
            }
            if (in_playback) {
                uicb_action_stop(widget);
            } else {
                uicb_action_play(widget);
            }
            return TRUE;
        }
        break;

      case GDK_Tab:
        if (!wave || in_playback) {
            break;
        }
        if (state == 0) {
            /* Swap the current and saved cursor/selection. */
            wave_swap_saved_position(wave);
            ui_update();
            return TRUE;
        }
        break;

      case GDK_Home:
        if (!wave || in_playback) {
            break;
        }
        if (state == GDK_CONTROL_MASK) {
            /* Move the cursor/selection to the beginning of the file. */
            wave_set_position(wave, 0, 1);
            ui_update();
            return TRUE;
        } else if (state == (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) {
            /* Select from the cursor position to the beginning of the file. */
            const uint64_t old_position = wave_position(wave);
            wave_set_selection(wave, 0, old_position, 1);
            ui_update();
            return TRUE;
        }
        break;

      case GDK_End:
        if (!wave || in_playback) {
            break;
        }
        if (state == GDK_CONTROL_MASK) {
            /* Move the cursor/selection to the end of the file. */
            wave_set_position(wave,
                              wave_samples(wave) - wave_select_len(wave), 1);
            ui_update();
            return TRUE;
        } else if (state == (GDK_CONTROL_MASK | GDK_SHIFT_MASK)) {
            /* Select from the cursor position to the end of the file. */
            wave_set_select_len(wave,
                                wave_samples(wave) - wave_position(wave), 1);
            ui_update();
            return TRUE;
        }
        break;

    }  // switch (event->keyval)

    return FALSE;
}

/*----------------------------------*/

gboolean uicb_event_button_press(GtkWidget *widget, GdkEventButton *event)
{
    /* Only react to the left button (button 1). */
    if (event->button != 1) {
        return FALSE;
    }

    /* Grab the focus, just so there's no annoying blinking caret in the
     * text fields.  (This callback is only used for the main canvas.) */
    gtk_widget_grab_focus(widget);

    /* Nothing else to do if there's no active wave. */
    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return FALSE;
    }

    /* Find the position associated with the clicked location. */
    GtkAllocation allocation;
    gtk_widget_get_allocation(widget, &allocation);
    const unsigned int width = allocation.width;
    const unsigned int x = (unsigned int)bound(event->x, 0, width-1);
    const uint64_t position = wave_samples(wave) * x / width;

    /* If we're playing audio, set the play position to the clicked
     * location (unless it's outside the play range, in which case we
     * ignore the event).  Otherwise, set the cursor position to the
     * clicked location (or select to the clicked location if it's a
     * shift-click), and cancel any selection. */
    if (in_playback) {
        if (position >= playback_start && position < playback_end) {
            playback_pos = position;
        }
    } else {
        if (event->state & GDK_SHIFT_MASK) {
            clicked_position = wave_position(wave);
            if (position < clicked_position) {
                wave_set_selection(wave, position,
                                   clicked_position - position, 1);
            } else {
                wave_set_selection(wave, clicked_position,
                                   position - clicked_position, 1);
            }
        } else {
            wave_set_selection(wave, position, 0, 1);
            /* Save the clicked position so dragging works properly. */
            clicked_position = wave_position(wave);
        }
    }

    ui_update();
    return TRUE;
}

/*----------------------------------*/

gboolean uicb_event_motion(GtkWidget *widget, GdkEventMotion *event)
{
    /* Only react to the left button (button 1). */
    if (!(event->state & GDK_BUTTON1_MASK)) {
        return FALSE;
    }

    /* Nothing to do if there's no active wave. */
    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return FALSE;
    }

    /* Don't react to events while playing audio. */
    if (in_playback) {
        return FALSE;
    }

    /* Adjust the size of the selection based on the current mouse position. */
    GtkAllocation allocation;
    gtk_widget_get_allocation(widget, &allocation);
    const unsigned int width = allocation.width;
    const unsigned int x = (unsigned int)bound(event->x, 0, width);
    const uint64_t new_position = wave_samples(wave) * x / width;
    if (new_position < clicked_position) {
        wave_set_selection(wave, new_position,
                           clicked_position - new_position, 1);
    } else {
        wave_set_selection(wave, clicked_position,
                           new_position - clicked_position, 1);
    }
    ui_update();

    return TRUE;
}

/*************************************************************************/

/**
 * uicb_update_*:  Callbacks for changes to position/selection/zoom values.
 *
 * [Parameters]
 *     widget: Relevant widget
 * [Return value]
 *     None
 */

void uicb_update_position(GtkWidget *widget)
{
    uint64_t value =
        (uint64_t)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));

    WaveFile *wave = manager_get_active_wave();
    if (wave) {
        if (value > wave_samples(wave) - wave_select_len(wave)) {
            value = wave_samples(wave) - wave_select_len(wave);
        }
        wave_set_position(wave, value, 1);
    }
    ui_update();
}

/*----------------------------------*/

void uicb_update_select_len(GtkWidget *widget)
{
    uint64_t value =
        (uint64_t)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));

    WaveFile *wave = manager_get_active_wave();
    if (wave) {
        if (value > wave_samples(wave) - wave_position(wave)) {
            value = wave_samples(wave) - wave_position(wave);
        }
        wave_set_select_len(wave, value, 1);
    }
    ui_update();
}

/*----------------------------------*/

void uicb_update_zoom(GtkWidget *widget)
{
    uint64_t value =
        (uint64_t)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget));

    WaveFile *wave = manager_get_active_wave();
    if (wave) {
        if (value < 1) {
            value = 1;
        } else if (value > wave_samples(wave)) {
            value = wave_samples(wave);
        }
        wave_set_zoom(wave, value);
    }
    ui_update();
}

/*************************************************************************/

/**
 * uicb_action_*:  Callbacks for actions triggered by menu items or toolbar
 * buttons.
 *
 * [Parameters]
 *     widget: Relevant widget
 * [Return value]
 *     None
 */

void uicb_action_new(GtkWidget *widget)
{
    unsigned int bits, channels, srate;

    if (get_properties("Select the sample data format to use for the new"
                       " file:", &bits, &channels, &srate)) {
        WaveFile *wave = wave_create(bits==8 ? FMT_U8 : FMT_S16,
                                     channels, srate);
        if (!wave) {
            ui_show_error("Failed to create new file.");
            return;
        }
        manager_add_wave(wave);
        manager_set_active_wave(wave);
    }
}

/*----------------------------------*/

void uicb_action_open(GtkWidget *widget)
{
    GtkWidget *dialog = ui_get_widget("open_dialog");
    const int result = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_hide(dialog);

    if (result == GTK_RESPONSE_OK) {
        char *path = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        WaveFile *wave = wave_load(path);
        if (!wave) {
            ui_show_error("Failed to open \"%s\".", path);
            g_free(path);
            return;
        }
        g_free(path);
        manager_add_wave(wave);
        manager_set_active_wave(wave);
    }
}

/*----------------------------------*/

void uicb_action_save(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave || !wave_path(wave) || !wave_writable(wave)) {
        return;
    }

    if (!wave_save(wave, wave_path(wave))) {
        ui_show_error("Failed to save \"%s\".", wave_path(wave));
    }

    ui_update();
}

/*----------------------------------*/

void uicb_action_save_as(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return;
    }

    GtkWidget *dialog = ui_get_widget("save_as_dialog");
    const int result = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_hide(dialog);

    if (result == GTK_RESPONSE_OK) {
        char *path = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
        if (!wave_save(wave, path)) {
            ui_show_error("Failed to save \"%s\".", path);
        }
        g_free(path);
        ui_update();
    }
}

/*----------------------------------*/

void uicb_action_close(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return;
    }

    ui_close_wave(wave);
}

/*----------------------------------*/

void uicb_action_quit(GtkWidget *widget)
{
    WaveFile *wave;
    while ((wave = manager_get_active_wave()) != NULL) {
        if (!ui_close_wave(wave)) {
            /* The user cancelled the close operation (or an error occurred
             * while saving changes), so don't quit. */
            return;
        }
    }

    gtk_main_quit();
}

/*----------------------------------*/

void uicb_action_cut(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave || wave_select_len(wave) == 0) {
        return;
    }

    /* Cut is just copy followed by delete.  However, if the copy fails
     * (probably due to lack of memory), we don't want to delete the
     * region, so we re-implement the copy action here. */

    // uicb_action_copy(widget);
    float *copied_data =
        wave_get_data(wave, wave_position(wave), wave_select_len(wave));
    if (!copied_data) {
        DMSG("Failed to copy data from wave file");
        return;
    }
    free(clipboard);
    clipboard          = copied_data;
    clipboard_length   = wave_select_len(wave);
    clipboard_format   = wave_format(wave);
    clipboard_channels = wave_channels(wave);
    clipboard_srate    = wave_srate(wave);

    uicb_action_delete(widget);
}

/*----------------------------------*/

void uicb_action_copy(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave || wave_select_len(wave) == 0) {
        return;
    }

    float *copied_data =
        wave_get_data(wave, wave_position(wave), wave_select_len(wave));
    if (!copied_data) {
        DMSG("Failed to copy data from wave file");
        return;
    }
    free(clipboard);
    clipboard          = copied_data;
    clipboard_length   = wave_select_len(wave);
    clipboard_format   = wave_format(wave);
    clipboard_channels = wave_channels(wave);
    clipboard_srate    = wave_srate(wave);

    /* We haven't changed any state related to the wave file, but we need
     * to enable the "Paste" and "Paste as New File" commands in case they
     * are still disabled (i.e. the clipboard was previously empty). */
    ui_update();
}

/*----------------------------------*/

void uicb_action_paste(GtkWidget *widget)
{
    if (!clipboard) {
        return;  // Nothing to paste!
    }

    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return;
    }

    /* Insert before delete, so we don't lose the data in the selection if
     * the insert fails. */
    const uint64_t position = wave_position(wave);
    const uint64_t select_len = wave_select_len(wave);
    if (!wave_insert_data(wave, position, clipboard, clipboard_length)) {
        DMSG("Failed to insert clipboard data into wave file");
        return;
    }
    if (wave_select_len(wave) > 0) {
        if (!wave_delete_data(wave, position + clipboard_length, select_len)) {
            DMSG("Failed to delete selection after pasting");
            return;
        }
    }

    /* Select the just-pasted data. */
    wave_set_selection(wave, position, clipboard_length, 1);

    ui_update();
}

/*----------------------------------*/

void uicb_action_paste_new(GtkWidget *widget)
{
    if (!clipboard) {
        return;  // Nothing to paste!
    }

    WaveFile *wave = wave_create(clipboard_format, clipboard_channels,
                                 clipboard_srate);
    if (!wave) {
        DMSG("Failed to create new wave file for paste");
        return;
    }

    if (!wave_insert_data(wave, wave_position(wave),
                          clipboard, clipboard_length)) {
        DMSG("Failed to insert clipboard data into new wave file");
        wave_destroy(wave);
        return;
    }
    /* Set both the current and saved position to the beginning of the file. */
    wave_set_selection(wave, 0, 0, 0);
    wave_swap_saved_position(wave);
    wave_set_selection(wave, 0, 0, 1);

    manager_add_wave(wave);
    manager_set_active_wave(wave);
}

/*----------------------------------*/

void uicb_action_delete(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave || wave_select_len(wave) == 0) {
        return;
    }

    if (!wave_delete_data(wave, wave_position(wave), wave_select_len(wave))) {
        DMSG("Failed to delete selection");
        return;
    }
    wave_set_select_len(wave, 0, 1);

    ui_update();
}

/*----------------------------------*/

void uicb_action_stop(GtkWidget *widget)
{
    if (in_playback) {
        audio_close();
        g_source_remove(playback_idle_id);
        in_playback = 0;
        ui_update();
    }
}

/*----------------------------------*/

void uicb_action_play(GtkWidget *widget)
{
    WaveFile *wave = manager_get_active_wave();
    if (!wave) {
        return;  // Nothing to play!
    }

    if (!in_playback) {
        playback_wave  = wave;
        playback_start = wave_select_len(wave) ? wave_position(wave) : 0;
        playback_end   = wave_select_len(wave)
                         ? wave_position(wave) + wave_select_len(wave)
                         : wave_samples(wave);
        playback_pos   = wave_position(wave) < wave_samples(wave)
                         ? wave_position(wave) : 0;
        if (playback_end > playback_start) {  // Sanity check
            /* wave_get_data() always returns 16-bit data, so we always
             * select 16-bit format for audio_open(). */
            if (audio_open(1, 16, wave_channels(wave), wave_srate(wave))) {
                in_playback = 1;
                playback_x = -1;
                playback_idle_id = g_idle_add((GSourceFunc)play_audio, NULL);
                ui_update();
            } else {
                ui_show_error("Failed to start playback.");
            }
        }
    }
}

/*----------------------------------*/

void uicb_action_record(GtkWidget *widget)
{
    if (in_playback) {
        return;
    }

    /* Ask the user for the sample format to use. */
    unsigned int bits, channels, srate;
    if (!get_properties("Select the sample data format to use for recording:",
                        &bits, &channels, &srate)) {
        return;
    }
    const WaveFormat format = bits==8 ? FMT_U8 : FMT_S16;

    /* Open the audio device for recording.  As with playback, we always
     * open the device in 16-bit mode regardless of the requested file
     * format. */
    if (!audio_open(0, 16, channels, srate)) {
        ui_show_error("Failed to start recording.");
        return;
    }

    /* Allocate a buffer for storing recorded data. */
    // FIXME: make the size user-selectable
    record_bufsize = 1<<27;  // 128M samples
    record_buffer = malloc(2 * channels * record_bufsize);
    if (!record_buffer) {
        ui_show_error("Failed to allocate a recording buffer.");
        audio_close();
        return;
    }

    /* Save the current active file, so we can switch back to it if
     * recording is subsequently cancelled. */
    record_last_wave = manager_get_active_wave();

    /* Create a new wave file to hold the data.  (We create it now rather
     * than after recording to guard against the possibility, however
     * remote, that creation fails and the user loses their recorded data.) */
    record_wave = wave_create(format, channels, srate);
    if (!record_wave) {
        ui_show_error("Failed to create a new file for recording.");
        free(record_buffer);
        record_buffer = NULL;
        audio_close();
        return;
    }
    manager_add_wave(record_wave);
    manager_set_active_wave(record_wave);
    ui_update();

    /* Start up the recording callback. */
    in_record = 1;
    recording = 0;
    record_low_l = record_high_l = 0;
    record_low_r = record_high_r = 0;
    record_idle_id = g_idle_add((GSourceFunc)record_audio, NULL);

    /* Activate the "Record" button, and deactivate "Stop".  (We leave them
     * both displayed rather than changing from one to the other in the
     * same location so an accidental double-click on the button doesn't
     * immediately stop the recording.) */
    gtk_widget_set_sensitive(ui_get_widget("recdialog_button_Record"), TRUE);
    gtk_widget_set_sensitive(ui_get_widget("recdialog_button_Stop"), FALSE);

    /* Display the recording dialog; show or hide the right channel's
     * monitor bar depending on how many channel's we're recording. */
    GtkWidget *dialog = ui_get_widget("record_dialog");
    gtk_widget_show_all(dialog);
    if (channels == 1) {
        gtk_widget_hide_all(ui_get_widget("recdialog_monitor_r_frame"));
    }

    /* We're done here; record_audio() will take care of everything else. */
}

/*----------------------------------*/

void uicb_action_about(GtkWidget *widget)
{
    GtkWidget *dialog = ui_get_widget("about_dialog");
    gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_hide(dialog);
}

/*************************************************************************/

/**
 * uicb_recdialog_cancel:  Callback for a "close" keypress (normally Escape)
 * on the recording dialog.  Also called by uicb_recdialog_action() for
 * cancel events.
 *
 * [Parameters]
 *     widget: Relevant widget (record_dialog)
 * [Return value]
 *     None
 */
void uicb_recdialog_cancel(GtkWidget *widget)
{
    audio_close();
    g_source_remove(record_idle_id);
    in_record = 0;

    free(record_buffer);
    record_buffer = NULL;

    manager_delete_wave(record_wave);
    wave_destroy(record_wave);
    record_wave = NULL;
    manager_set_active_wave(record_last_wave);

    gtk_widget_hide(widget);
    ui_update();
}

/*----------------------------------*/

/**
 * uicb_recdialog_action:  Callback for button presses on the recording
 * dialog.
 *
 * [Parameters]
 *          widget: Relevant widget (record_dialog)
 *     response_id: Response ID of the pressed button
 * [Return value]
 *     None
 */
void uicb_recdialog_action(GtkWidget *widget, gint response_id)
{
    /* On a window close event, cancel the dialog if we haven't started
     * recording yet; otherwise save the recorded data as though Stop was
     * pressed. */
    if (response_id == GTK_RESPONSE_DELETE_EVENT) {
        if (!recording) {
            response_id = GTK_RESPONSE_CANCEL;
        } else {
            response_id = 2;  // i.e. Stop
        }
    }

    switch (response_id) {

      case GTK_RESPONSE_CANCEL:
        uicb_recdialog_cancel(widget);
        break;

      case 1:  // Record
        gtk_widget_set_sensitive(ui_get_widget("recdialog_button_Record"), FALSE);
        gtk_widget_set_sensitive(ui_get_widget("recdialog_button_Stop"), TRUE);

        record_length = 0;
        record_overruns = 0;
        record_clipped = 0;
        recording = 1;

        break;

      case 2:  // Stop

        audio_close();
        g_source_remove(record_idle_id);
        in_record = 0;

        void *ptr = realloc(record_buffer,
                            4 * wave_channels(record_wave) * record_length);
        if (!ptr) {  // Should never happen (we're shrinking), but just in case
            ptr = record_buffer;
        }
        wave_take_data(record_wave, ptr, record_length);
        record_buffer = NULL;

        gtk_widget_hide(widget);
        ui_update();
        break;

    }  // switch (response_id)
}

/*----------------------------------*/

/**
 * uicb_recdialog_monitor_expose:  Callbacks for expose events on the
 * recording dialog's audio monitor bars.
 *
 * [Parameters]
 *     widget: Relevant widget (audio monitor drawing area)
 *      event: Event data
 * [Return value]
 *     None
 */
gboolean uicb_recdialog_monitor_expose(GtkWidget *widget,
                                       GdkEventExpose *event)
{
    /* Create a graphics context for drawing. */
    GdkDrawable *drawable = GDK_DRAWABLE(gtk_widget_get_window(widget));
    GdkGC *gc = gdk_gc_new(drawable);
    if (!gc) {
        DMSG("Failed to create GC");
        return FALSE;
    }

    /* Retrieve the size of the drawing area. */
    int width, height;
    gdk_drawable_get_size(drawable, &width, &height);

    /* Clear the widget to the background color. */
    gtk_style_apply_default_background(gtk_widget_get_style(widget),
                                       GDK_WINDOW(drawable), TRUE,
                                       GTK_STATE_NORMAL, &event->area,
                                       0, 0, width, height);

    /* Retrieve the low and high values and determine the drawing color
     * for this channel.  Since Glade doesn't know how to send arbitrary
     * arguments to signal handlers, we check whether the widget itself
     * is the left or right channel widget. */
    unsigned int low, high;  // 0x0000..0xFFFF
    GdkColor color;
    memset(&color, 0, sizeof(color));
    if (widget == ui_get_widget("recdialog_monitor_r")) {
        low   = (unsigned int)(bound((int)roundf(record_low_r*32768), -32768, 32767) + 0x8000);
        high  = (unsigned int)(bound((int)roundf(record_high_r*32768), -32768, 32767) + 0x8000);
        color.pixel = 0xFFFF0000;
    } else {
        low   = (unsigned int)(bound((int)roundf(record_low_l*32768), -32768, 32767) + 0x8000);
        high  = (unsigned int)(bound((int)roundf(record_high_l*32768), -32768, 32767) + 0x8000);
        color.pixel = 0xFF0000FF;
    }

    /* Draw a colored bar covering the range [low,high]. */
    const unsigned int xlow  = (low  * width) / 0x10000;
    const unsigned int xhigh = (high * width) / 0x10000;
    gdk_gc_set_foreground(gc, &color);
    gdk_draw_rectangle(drawable, gc, TRUE, xlow, 0, xhigh - xlow + 1, height);

    /* Destroy the graphics context we created. */
    gdk_gc_destroy(gc);

    return TRUE;
}

/*************************************************************************/
/************************* Other local functions *************************/
/*************************************************************************/

/**
 * get_properties:  Display the sample properties dialog and allow the user
 * to select sample resolution, number of channels, and sampling rate.  The
 * variables pointed to by the pointer arguments should hold the desired
 * initial values for each parameter, and will be updated on successful
 * return to hold the values selected by the user.
 *
 * [Parameters]
 *           prompt: Prompt message to display in dialog
 *         bits_ptr: Pointer to variable holding sample resolution
 *     channels_ptr: Pointer to variable holding number of channels
 *        srate_ptr: Pointer to variable holding sampling rate
 * [Return value]
 *     True (nonzero) if the user selected the "OK" button on the dialog,
 *     otherwise false (zero)
 */
static int get_properties(const char *prompt, unsigned int *bits_ptr,
                          unsigned int *channels_ptr, unsigned int *srate_ptr)
{
    PRECOND(prompt != NULL, return 0);
    PRECOND(bits_ptr != NULL, return 0);
    PRECOND(channels_ptr != NULL, return 0);
    PRECOND(srate_ptr != NULL, return 0);

    #define SET_ACTIVE(name) \
        gtk_toggle_button_set_active( \
            GTK_TOGGLE_BUTTON(ui_get_widget("properties_" #name)), TRUE)
    #define IS_ACTIVE(name) \
        gtk_toggle_button_get_active( \
            GTK_TOGGLE_BUTTON(ui_get_widget("properties_" #name)))

    GtkWidget *dialog = ui_get_widget("properties_dialog");
    gtk_label_set_text(GTK_LABEL(ui_get_widget("properties_prompt")), prompt);

    /* If there's an active wave file, use its parameters as the defaults;
     * otherwise leave the state of the parameter widgets alone. */
    WaveFile *wave = manager_get_active_wave();
    if (wave) {
        const WaveFormat format = wave_format(wave);
        const unsigned int channels = wave_channels(wave);
        const unsigned int srate = wave_srate(wave);
        if (format == FMT_U8) {
            SET_ACTIVE(bits_8);
        } else {
            SET_ACTIVE(bits_16);
        }
        if (channels == 1) {
            SET_ACTIVE(channels_1);
        } else {
            SET_ACTIVE(channels_2);
        }
        if (srate == 48000) {
            SET_ACTIVE(srate_48000);
        } else if (srate == 24000) {
            SET_ACTIVE(srate_24000);
        } else if (srate == 44100) {
            SET_ACTIVE(srate_44100);
        } else if (srate == 22050) {
            SET_ACTIVE(srate_22050);
        } else if (srate == 32000) {
            SET_ACTIVE(srate_32000);
        } else if (srate == 16000) {
            SET_ACTIVE(srate_16000);
        } else {
            SET_ACTIVE(srate_other);
            gtk_spin_button_set_value(GTK_SPIN_BUTTON(ui_get_widget("properties_srate_entry")), srate);
        }
    }  // if (wave)

    /* Actually show the dialog and wait for user action. */
    const int result = gtk_dialog_run(GTK_DIALOG(dialog));
    gtk_widget_hide(dialog);
    if (result != GTK_RESPONSE_OK) {
        return 0;
    }

    /* Check the radio buttons to find out what the user chose. */
    if (IS_ACTIVE(bits_8)) {
        *bits_ptr = 8;
    } else {
        *bits_ptr = 16;
    }
    if (IS_ACTIVE(channels_1)) {
        *channels_ptr = 1;
    } else {
        *channels_ptr = 2;
    }
    if (IS_ACTIVE(srate_48000)) {
        *srate_ptr = 48000;
    } else if (IS_ACTIVE(srate_24000)) {
        *srate_ptr = 24000;
    } else if (IS_ACTIVE(srate_44100)) {
        *srate_ptr = 44100;
    } else if (IS_ACTIVE(srate_22050)) {
        *srate_ptr = 22050;
    } else if (IS_ACTIVE(srate_32000)) {
        *srate_ptr = 32000;
    } else if (IS_ACTIVE(srate_16000)) {
        *srate_ptr = 16000;
    } else {
        *srate_ptr = gtk_spin_button_get_value(GTK_SPIN_BUTTON(ui_get_widget("properties_srate_entry")));
    }

    return 1;

    #undef SET_ACTIVE
    #undef IS_ACTIVE
}

/*************************************************************************/

/**
 * play_audio:  Play the next block of audio, terminating playback if the
 * end of the playback range is reached.  Called as an idle function from
 * the GTK+ main loop.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     TRUE if the function should be called again, else FALSE
 */
static gboolean play_audio(void)
{
    PRECOND(in_playback, return FALSE);
    const char *error = NULL;

    if (playback_pos < playback_end && audio_playable()) {
        const unsigned int len =
            ubound(playback_end - playback_pos, audio_buffer_size());
        float *data = wave_get_data(playback_wave, playback_pos, len);
        if (!data) {
            error = "Unable to access the file's data.";
            audio_close();
            in_playback = 0;
        } else {
            int16_t *data16 = (int16_t *)data;
            const int channels = wave_channels(playback_wave);
            for (unsigned int i = 0; i < len * channels; i++) {
                data16[i] = bound((int)roundf(data[i]*32768), -32768, 32767);
            }
            if (!audio_play(data, len)) {
                error = "Unable to play the audio data.";
                audio_close();
                in_playback = 0;
            } else {
                playback_pos += len;
            }
        }
        free(data);
    } else {
        nanosleep(&((const struct timespec){.tv_sec = 0, .tv_nsec = 10000000}),
                  NULL);
    }

    GDK_THREADS_ENTER();

    if (in_playback && playback_pos >= playback_end) {
        const int loop_enabled = gtk_toggle_tool_button_get_active(
            GTK_TOGGLE_TOOL_BUTTON(ui_get_widget("button_Loop"))
        );
        if (loop_enabled) {
            playback_pos = playback_start;
        } else if (audio_samples_buffered() == 0) {
            audio_flush();  // Just in case
            audio_close();
            g_source_remove(playback_idle_id);
            in_playback = 0;
        }
    }

    if (in_playback) {

        /* Rather than calling ui_update(), which triggers a full redraw of
         * the main window, just update the position display ourselves and
         * only redraw the necessary portion of the waveform display. */

        uint64_t real_position;
        const uint64_t buffered = audio_samples_buffered();
        if (buffered > playback_pos - playback_start) {
            /* We just looped, so show the proper position at the end of
             * the playback region. */
            real_position =
                playback_end - (buffered - (playback_pos - playback_start));
        } else {
            real_position = playback_pos - buffered;
        }

        GtkWidget *position_widget = ui_get_widget("position");
        g_signal_handlers_block_by_func(position_widget,
                                        uicb_update_position, NULL);
        gtk_spin_button_set_value(GTK_SPIN_BUTTON(position_widget),
                                  real_position);
        g_signal_handlers_unblock_by_func(position_widget,
                                          uicb_update_position, NULL);

        GtkWidget *main_canvas = ui_get_widget("main_canvas");
        GtkAllocation allocation;
        gtk_widget_get_allocation(main_canvas, &allocation);
        const unsigned int last_x =
            (playback_x < 0) ? 0 : (unsigned int)playback_x;
        playback_x =
            ubound(real_position * allocation.width
                   / wave_samples(playback_wave), allocation.width - 1);
        allocation.x = min(playback_x, last_x);
        allocation.y = 0;
        allocation.width = max(playback_x, last_x) - allocation.x + 1;
        gdk_window_invalidate_rect(gtk_widget_get_window(main_canvas),
                                   &allocation, FALSE);

    } else {  // !in_playback

        /* Playback ended, so we should do a full update to re-enable all
         * the widgets and so on. */
        ui_update();

        /* Also display an error message if requested. */
        if (error) {
            ui_show_error("%s", error);
        }

    }

    GDK_THREADS_LEAVE();

    return in_playback;
}

/*************************************************************************/

/**
 * record_audio:  Read a block of audio from the audio device, and update
 * the audio monitor display on the recording dialog.  Called as an idle
 * function from the GTK+ main loop.
 *
 * [Parameters]
 *     None
 * [Return value]
 *     TRUE if the function should be called again, else FALSE
 */
static gboolean record_audio(void)
{
    PRECOND(in_record, return FALSE);
    PRECOND(record_wave != NULL, in_record = 0; return FALSE);

    /* Low and high points in this segment of audio.  high < low is used
     * to mean "don't draw" (e.g. because there was an audio overrun and
     * we didn't get any data). */
    float low_l = 1, high_l = -1;
    float low_r = 1, high_r = -1;

    /* Read a reasonably-sized (large enough to minimize system call
     * overhead, but small enough not to slow down GTK while we wait)
     * chunk of audio and process it. */
    const unsigned int channels = wave_channels(record_wave);
    const unsigned int samples_to_read = 512;  // FIXME: good enough for now
    int16_t buffer[channels * samples_to_read];
    if (audio_record(buffer, samples_to_read)) {
        const int16_t *ptr = buffer;
        for (unsigned int i = 0; i < samples_to_read; i++, ptr += channels) {
            float sample = *ptr / 32768.0f;
            low_l  = min(low_l,  sample);
            high_l = max(high_l, sample);
            if (channels == 2) {
                sample = ptr[1] / 32768.0f;
                low_r  = min(low_r,  sample);
                high_r = max(high_r, sample);
            }
        }
        if (recording) {
            unsigned int to_copy = ubound(samples_to_read,
                                          record_bufsize - record_length);
            if (to_copy > 0) {
                for (unsigned int i = 0; i < channels * to_copy; i++) {
                    record_buffer[channels * record_length + i] =
                        buffer[i] / 32768.0f;
                }
                record_length += to_copy;
            }
        }
    } else if (errno == EPIPE) {
        if (recording) {
            record_overruns++;
        }
    } else {
        ui_show_error("Failed to read audio data.%s",
                      recording ? "  The recording will now be stopped." : "");
        /* Conveniently, faking a DELETE_EVENT response will get exactly
         * the action we want (cancel if monitoring, stop if recording). */
        GDK_THREADS_ENTER();
        uicb_recdialog_action(ui_get_widget("record_dialog"),
                              GTK_RESPONSE_DELETE_EVENT);
        GDK_THREADS_LEAVE();
        return FALSE;
    }

    /* Check for possible clipping.  (We can't detect whether there was
     * actual clipping, of course, but a value at the very edge of the
     * range is a danger sign.  Note that depending on the hardware
     * involved and the current capture settings, it's still possible to
     * get clipping while staying well within these limits.) */
    if (low_l <= -1 || high_l >= 0x7FFF/32768.0f
     || low_r <= -1 || high_r >= 0x7FFF/32768.0f) {
        if (recording) {
            record_clipped = 1;
        }
    }

    GDK_THREADS_ENTER();

    /* Update the audio monitor display, if appropriate. */
    if (high_l >= low_l) {
        record_low_l  = low_l;
        record_high_l = high_l;
        GtkWidget *monitor = ui_get_widget("recdialog_monitor_l");
        GtkAllocation allocation;
        gtk_widget_get_allocation(monitor, &allocation);
        allocation.x = allocation.y = 0;
        gdk_window_invalidate_rect(gtk_widget_get_window(monitor),
                                   &allocation, FALSE);
    }
    if (high_r >= low_r) {
        record_high_r = high_r;
        record_low_r  = low_r;
        GtkWidget *monitor = ui_get_widget("recdialog_monitor_r");
        GtkAllocation allocation;
        gtk_widget_get_allocation(monitor, &allocation);
        allocation.x = allocation.y = 0;
        gdk_window_invalidate_rect(gtk_widget_get_window(monitor),
                                   &allocation, FALSE);
    }

    /* Update the status line. */
    GtkLabel *status_label = GTK_LABEL(ui_get_widget("recdialog_status"));
    if (recording) {
        char status_buf[1000];
        snprintf(status_buf, sizeof(status_buf), "%.2f s (%llu kB) recorded",
                 (double)record_length / (double)wave_srate(record_wave),
                 (unsigned long long)((2 * wave_channels(record_wave)
                                         * record_length) + 44) / 1000);
        if (record_clipped) {
            snprintf(status_buf + strlen(status_buf),
                     sizeof(status_buf) - strlen(status_buf),
                     ", <i>possible clipping</i>");
        }
        if (record_overruns > 0) {
            snprintf(status_buf + strlen(status_buf),
                     sizeof(status_buf) - strlen(status_buf),
                     ", <b>%u overrun%s</b>",
                     record_overruns, record_overruns==1 ? "" : "s");
        }
        gtk_label_set_markup(status_label, status_buf);
    } else {
        gtk_label_set_text(status_label,
                           "Select \"Record\" to start recording.");
    }

    GDK_THREADS_LEAVE();

    return TRUE;
}

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
