/*
 * XWaveN -- a simple audio file editor
 * Written by Andrew Church <achurch@achurch.org>
 *
 * src/wave.h: Interface for using and editing individual wave files.
 */

#ifndef WAVE_H
#define WAVE_H

#include <gdk/gdkcolor.h>

/*************************************************************************/

/**
 * WaveFormat: Constants representing various data formats.  All multi-byte
 * formats are stored in little-endian byte order in RIFF WAVE files.
 */
typedef enum WaveFormat {
    FMT_U8 = 1, // 8-bit unsigned integer
    FMT_S16,    // 16-bit signed integer
    FMT_S24,    // 24-bit signed integer
    FMT_F32,    // 32-bit floating point
} WaveFormat;

/**
 * wave_create:  Create a new, empty wave file.
 *
 * [Parameters]
 *       format: Format of the wave data (FMT_*)
 *     channels: Number of channels in the wave file (1-6)
 *        srate: Sampling rate of the wave data (at least 1)
 * [Return value]
 *     Newly-created wave file, or NULL on error
 */
extern WaveFile *wave_create(WaveFormat format, unsigned int channels,
                             unsigned int srate);

/**
 * wave_load:  Load a .WAV file from disk, returning the associated wave
 * file structure.
 *
 * [Parameters]
 *     path: Pathname of file to load
 * [Return value]
 *     Loaded wave file, or NULL on error
 */
extern WaveFile *wave_load(const char *path);

/**
 * wave_save:  Save a wave file to disk.  The newly-created file will be
 * used for subsequent access.
 *
 * [Parameters]
 *     wave: Wave file to save
 *     path: New pathname for file
 * [Return value]
 *     Nonzero on success, zero on error
 */
extern int wave_save(WaveFile *wave, const char *path);

/**
 * wave_destroy:  Destroy a wave file and all associated resources.  Does
 * nothing if wave == NULL.
 *
 * [Parameters]
 *     wave: Wave file to destroy
 * [Return value]
 *     None
 */
extern void wave_destroy(WaveFile *wave);

/*----------------------------------*/

/**
 * wave_path:  Return the pathname associated with the given wave file, or
 * NULL if the wave file does not have an associated pathname.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Associated pathname, or NULL if none
 */
extern const char *wave_path(const WaveFile *wave);

/**
 * wave_writable:  Return whether the disk file backing the wave data for
 * the given wave file is writable.  Always returns true for wave files
 * which are not backed by a disk file, such as newly-created empty files.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     True (nonzero) if the disk file is writable or there is no
 *     associated disk file, else false (zero)
 */
extern int wave_writable(const WaveFile *wave);

/**
 * wave_modified:  Returns whether the given wave file has been modified
 * since it was last loaded or saved.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     True (nonzero) if the wave file has been modified since last loaded
 *     or saved, else false (zero)
 */
extern int wave_modified(const WaveFile *wave);

/**
 * wave_format, wave_bits, wave_channels, wave_srate, wave_samples:  Return
 * the format, bit resolution, number of channels, sampling rate, or number
 * of samples, respectively, for the given wave file.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Relevant wave file parameter
 */
extern WaveFormat wave_format(const WaveFile *wave);
extern unsigned int wave_bits(const WaveFile *wave);
extern unsigned int wave_channels(const WaveFile *wave);
extern unsigned int wave_srate(const WaveFile *wave);
extern uint64_t wave_samples(const WaveFile *wave);

/**
 * wave_position, wave_select_len, wave_zoom:  Return the current cursor
 * position, selection length, or zoom factor, respectively for the given
 * wave file.  All return values are in units of samples.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Relevant wave file parameter
 */
extern uint64_t wave_position(const WaveFile *wave);
extern uint64_t wave_select_len(const WaveFile *wave);
extern uint64_t wave_zoom(const WaveFile *wave);

/**
 * wave_canvas_width:  Return the width of the wave file's canvas window,
 * in pixels.
 *
 * [Parameters]
 *     wave: Wave file from which to return data
 * [Return value]
 *     Width of canvas window
 */
extern unsigned int wave_canvas_width(const WaveFile *wave);

/*----------------------------------*/

/**
 * wave_set_position, wave_set_select_len, wave_set_zoom:  Set the current
 * cursor position, selection length, or zoom factor, respectively for the
 * given wave file.  All values are in units of samples.
 *
 * [Parameters]
 *            wave: Wave file from which to return data
 *           value: New value for relevant wave file parameter
 *     move_canvas: Nonzero to cause the wave file's canvas window to be
 *                     moved based on the new value, else zero (only for
 *                     set_position() and set_select_len())
 * [Return value]
 *     None
 */
extern void wave_set_position(WaveFile *wave, uint64_t value, int move_canvas);
extern void wave_set_select_len(WaveFile *wave, uint64_t value, int move_canvas);
extern void wave_set_zoom(WaveFile *wave, uint64_t value);

/**
 * wave_set_selection:  Set the current cursor position and selection
 * length for the given wave file.  This is a convenience function that
 * wraps set_position() and set_select_len(), and relieves callers from
 * having to worry about the proper order in which to set the values so
 * the selection does not run over the end of the file.
 *
 * [Parameters]
 *            wave: Wave file from which to return data
 *        position: New cursor position, in samples
 *      select_len: New selection length, in samples
 *     move_canvas: Nonzero to cause the wave file's canvas window to be
 *                     moved based on the new values, else zero
 * [Return value]
 *     None
 */
extern void wave_set_selection(WaveFile *wave, uint64_t position,
                               uint64_t select_len, int move_canvas);

/**
 * wave_swap_saved_position:  Swap the current and saved positions for the
 * given wave file.
 *
 * [Parameters]
 *     wave: Wave file for which to swap positions
 * [Return value]
 *     None
 */
extern void wave_swap_saved_position(WaveFile *wave);

/*----------------------------------*/

/**
 * wave_get_data:  Return sample data from the specified portion of the
 * wave file.
 *
 * [Parameters]
 *       wave: Wave file from which to return data
 *      start: Index of first sample to return
 *     length: Number of samples to return
 * [Return value]
 *     Sample data as floating-point values in sample-major, channel-minor
 *     order, or NULL on error
 */
extern float *wave_get_data(const WaveFile *wave, uint64_t start, uint64_t length);

/**
 * wave_insert_data:  Insert the given data into the given wave file.  The
 * data is assumed to have the same number of channels as the wave file.
 *
 * [Parameters]
 *         wave: Wave file into which to insert data
 *     position: Position at which to insert data, in samples
 *         data: Data to insert, as floating-point values
 *       length: Length of data to insert, in samples
 * [Return value]
 *     Nonzero on success, zero on error
 */
extern int wave_insert_data(WaveFile *wave, uint64_t position,
                            const float *data, uint64_t length);

/**
 * wave_delete_data:  Delete the given section of the given wave file.
 *
 * [Parameters]
 *         wave: Wave file from which to delete data
 *     position: Position at which to delete data, in samples
 *       length: Length of data to delete, in samples
 * [Return value]
 *     Nonzero on success, zero on error
 */
extern int wave_delete_data(WaveFile *wave, uint64_t position, uint64_t length);

/**
 * wave_take_data:  Replace the entire contents of the given wave file with
 * the given data.  The data is assumed to have the same number of channels
 * as the wave file.  This routine always succeeds if its preconditions are
 * fulfilled.
 *
 * After calling this function, the data buffer passed to this routine is
 * owned by the wave file, and must not be modified or freed by the caller.
 *
 * [Parameters]
 *       wave: Wave file into which to store data
 *       data: Data to store, as floating-point values
 *     length: Length of data to store, in samples
 * [Return value]
 *     None
 */
extern void wave_take_data(WaveFile *wave, float *data, uint64_t length);

/*----------------------------------*/

/**
 * wave_get_full_canvas:  Return data for drawing the entire waveform on a
 * canvas of the given width.  The data will be cached, so subsequent
 * requests for the same width will not require re-reading the data from
 * disk.
 *
 * The returned data is a sequence of unsigned 16-bit values, representing
 * the lowest and highest sample values for each channel in the range of
 * samples corresponding to each pixel:
 *    pixel0_ch0_low, pixel0_ch0_high, pixel0_ch1_low, ..., pixel1_ch0_low, ...
 * For each value, 0x0000 corresponds to the lowest possible sample value
 * (e.g. -0x8000 for signed 16-bit data), and 0xFFFF corresponds to the
 * highest possible sample value (e.g. +0x7FFF for signed 16-bit data).
 *
 * [Parameters]
 *      wave: Wave file for which to return data
 *     width: Width of canvas for which to return data
 * [Return value]
 *     Canvas drawing data, or NULL on error
 */
extern const uint16_t *wave_get_full_canvas(WaveFile *wave, unsigned int width);

/**
 * wave_get_channel_color:  Return a GdkColor for a channel in the given
 * waveform.  Colors are selected so that the same color represents the
 * same spatial position (left, center, LFE, and so on).
 *
 * [Parameters]
 *        wave: Wave file for which to return a color
 *     channel: Channel index (zero-based)
 * [Return value]
 *     Color for the given channel
 */
extern const GdkColor *wave_get_channel_color(const WaveFile *wave, int channel);

/*************************************************************************/

#endif  // WAVE_H

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
